from astroid import nodes  # type: ignore
from astroid.const import Context
from typing import TYPE_CHECKING, Optional, Tuple, Union, List, Any, Dict, Set
from enum import Enum

from pylint.checkers import BaseChecker  # type: ignore
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import (
    get_name,
    get_assigned_to,
    is_any_assign,
    is_pure_builtin,
    get_const_value,
)

ExprRepresentation = str
Comparison = str
Value = Any
NodeCmpValue = Optional[Tuple[nodes.NodeNG, Comparison, Value]]
SameDirComparisons = Dict[
    ExprRepresentation, Dict[Comparison, Tuple[List[ExprRepresentation], List[bool]]]
]
# When you take a key (1), then one key (cmp) from the corresponding dictionary and an element from the
# first array from its corresponding value (2), you get a comparison ((1) (cmp) (2)). The List[bool]
# indicates, whether the comparison in the original source code is in the same direction as in the dictionary.


def is_modulo_residue(value: Any, modulo: int) -> bool:
    return isinstance(value, int) and 0 <= value < modulo


class Comparisons(Enum):
    EQUALITY = 0
    INEQUALITY = 1
    OTHER = 2


# I use this to represent all the pairs of type 'x % 2 == 0 and y % 2 != 1'
# expression_pair is the ('x', 'y'), modulo is 2, comparisons_for_equality is {},
# comparisons_for_inequality is [{0}, {}],
# comparisons_for_true is [{0}, {}] and
# comparisons_pairs is [('x % 2 == 0 and y % 2 != 1', Comparisons.OTHER)]
# Note that we have some redundancy here, but this makes the code easier later and
# this is not some code that would be run all the time, in fact it is quite rare.
class PairsOfModComparisonsWithNum:
    def __init__(
        self,
        expr1: ExprRepresentation,
        cmp1: Comparison,
        value1: Value,
        expr2: ExprRepresentation,
        cmp2: Comparison,
        value2: Value,
        modulo: int,
        comparison_pair_string: str,
    ) -> None:
        # the two expressions that have modulo applied to them
        self.expression_pair = (expr1, expr2)
        self.modulo = modulo

        # for expressions of type 'x % n == a and y % n == a' we add to this set the 'a'
        self.comparisons_for_equality: Set[int] = set()

        # stores expressions like 'expr1 % modulo == a and expr2 % modulo != b',
        # index represents the number that 'expr1 % modulo' is equal to and the set is
        # the set of all the values 'expr2 % modulo' could be equal to.
        self.comparisons_for_inequality: List[Set[int]] = [set() for _ in range(modulo)]
        self.comparisons_for_true: List[Set[int]] = [set() for _ in range(modulo)]

        # this is here so that we can easily get the original string
        self.comparisons_pairs: List[Tuple[str, Comparisons]] = []

        self.update(cmp1, value1, cmp2, value2, comparison_pair_string)

    def update_comparisons(
        self,
        comparisons: List[Set[int]],
        cmp1: Comparison,
        value1: Value,
        cmp2: Comparison,
        value2: Value,
    ) -> None:
        for v1 in self.get_possible_values_mod(cmp1, value1):
            comparisons[v1].update(self.get_possible_values_mod(cmp2, value2))

    def get_possible_values_mod(self, cmp1: Comparison, value1: Value) -> List[int]:
        # Note that here we use that for example 'x % 3 != 1' can be rewritten to 'x % 3 == 0 or x % 3 == 2'
        if cmp1 == "==":
            return [value1]

        return [i for i in range(self.modulo) if i != value1]

    def update(
        self,
        cmp1: Comparison,
        value1: Value,
        cmp2: Comparison,
        value2: Value,
        comparison_pair_string: str,
    ) -> None:
        if cmp1 == "==" and cmp2 == "==" and value1 == value2:
            self.comparisons_for_equality.add(value1)
            self.comparisons_pairs.append((comparison_pair_string, Comparisons.EQUALITY))
        elif (cmp1 == "==" and cmp2 == "==") or (
            (cmp1 != "!=" or cmp2 != "!=") and value1 == value2
        ):
            self.update_comparisons(self.comparisons_for_inequality, cmp1, value1, cmp2, value2)
            self.comparisons_pairs.append((comparison_pair_string, Comparisons.INEQUALITY))
        else:
            self.comparisons_pairs.append((comparison_pair_string, Comparisons.OTHER))

        self.update_comparisons(self.comparisons_for_true, cmp1, value1, cmp2, value2)

    def always_true(self) -> bool:
        for comparisons in self.comparisons_for_true:
            if len(comparisons) != self.modulo:
                return False

        return True

    def has_part_simplifiable_to_equality(self) -> bool:
        return len(self.comparisons_for_equality) == self.modulo

    def has_part_simplifiable_to_inequality(self) -> bool:
        for comparisons in self.comparisons_for_inequality:
            if len(comparisons) != self.modulo - 1:
                return False

        return True


class SimplifiableIf(BaseChecker):  # type: ignore
    name = "simplifiable-if"
    msgs = {
        "R6201": (
            "The if statement can be replaced with 'return %s'",
            "simplifiable-if-return",
            "Emitted when the condition of an if statement can be returned directly (possibly negated).",
        ),
        "R6202": (
            "The if statement can be replaced with 'return %s'",
            "simplifiable-if-return-conj",
            "Emitted when the condition of an if statement and the returned values "
            "can be combined using logical operators.",
        ),
        "R6203": (
            "The conditional assignment can be replaced with '%s = %s'",
            "simplifiable-if-assignment",
            "Emitted when the condition of an if statement can be assigned directly (possibly negated).",
        ),
        "R6210": (
            "The conditional assignment can be replaced with '%s = %s'",
            "simplifiable-if-assignment-conj",
            "Emitted when the condition of an if statement and the assigned values "
            "can be combined using logical operators.",
        ),
        "R6204": (
            "The if expression can be replaced with '%s'",
            "simplifiable-if-expr",
            "Emitted when the condition of an if expression can be returned directly (possibly negated).",
        ),
        "R6209": (
            "The if expression can be replaced with '%s'",
            "simplifiable-if-expr-conj",
            "Emitted when the condition of an if expression and the returned values "
            "can be combined using logical operators.",
        ),
        "R6205": (
            "Use 'if %s: <else body>' instead of 'pass'",
            "simplifiable-if-pass",
            "Emitted when there is an if condition with a pass in the positive branch.",
        ),
        "R6206": (
            "Both branches should return a value explicitly (one returns implicit None)",
            "no-value-in-one-branch-return",
            "Emitted when one branch returns a value and the other just returns.",
        ),
        "R6207": (
            "The if statement can be merged with the nested one to 'if %s:'",
            "simplifiable-if-nested",
            "Emitted when the condition of an if statement can be merged "
            "with its nested if's condition using logical operators.",
        ),
        "R6208": (
            "The if statement can be merged with the following one to 'if %s:'",
            "simplifiable-if-seq",
            "Emitted when the condition of an if statement can be merged "
            "with the next if's condition using logical operators.",
        ),
        "R6211": (
            "'%s' can be replaced with '%s'",
            "simplifiable-with-abs",
            "Emitted when there is a problem like x < 4 and x > -4 and suggests abs(x) < 4.",
        ),
        "R6212": (
            "'%s' can be simplified to '%s'. Simplify the condition if it is on purpose, or change it if it was not.",
            "redundant-compare-in-condition",
            "Emitted when there is a problem like x > 4 or x > 3 and suggests x > 3. (ie min{4, 3})",
        ),
        "R6213": (
            "'%s' can be replaced with '%s'",
            "redundant-compare-avoidable-with-max-min",
            "Emitted when there is a problem like 'x > a and x > b' and suggests x > max(a, b).",
        ),
        "R6214": (
            "'%s' can be replaced with '%s'",
            "using-compare-instead-of-equal",
            "Emitted when there is a problem like x >= 0 and x <= 0 and suggests x == 0.",
        ),
        "R6215": (
            "'%s' can be replaced with '%s'",
            "simplifiable-test-by-equals",
            "Emitted when there is a problem like 'x % 2 == 0 and y % 2 == 0 or x % 2 == 1 and y % 2 == 1' and suggests 'x % 2 == y % 2'.",
        ),
    }

    def _is_bool(self, node: nodes.NodeNG) -> bool:
        return isinstance(node, nodes.Const) and node.pytype() == "builtins.bool"

    def _simplifiable_if_message(
        self, node: nodes.If, then: nodes.NodeNG, new_cond: str, only_replaces: bool
    ) -> None:
        extra = "" if only_replaces else "-conj"
        if isinstance(node, nodes.IfExp):
            self.add_message("simplifiable-if-expr" + extra, node=node, args=(new_cond))
        elif isinstance(then, nodes.Return):
            self.add_message("simplifiable-if-return" + extra, node=node, args=(new_cond))
        else:
            self.add_message(
                "simplifiable-if-assignment" + extra,
                node=node,
                args=(get_name(get_assigned_to(then)[0]), new_cond),
            )

    def _get_refactored(self, *args: Union[str, nodes.NodeNG]) -> str:
        result = []
        i = 0
        while i < len(args):
            arg = args[i]
            if isinstance(arg, str) and arg == "not":
                v = args[i + 1]
                assert isinstance(v, nodes.NodeNG)
                refactored = v.as_string()
                if isinstance(v, nodes.BoolOp):
                    result.append(f"<negated ({refactored})>")
                else:
                    result.append(f"<negated {refactored}>")
                i += 2
            elif isinstance(arg, str) and arg in ("and", "or"):
                prev = args[i - 1]
                next_ = args[i + 1]
                if (
                    isinstance(prev, nodes.BoolOp)
                    and prev.op != arg
                    and (i - 2 < 0 or args[i - 2] != "not")
                ):
                    result[-1] = f"({result[-1]})"
                result.append(arg)
                i += 1
                if isinstance(next_, nodes.BoolOp) and next_.op != arg:
                    result.append(f"({next_.as_string()})")
                    i += 1
            elif isinstance(arg, nodes.NodeNG):
                result.append(arg.as_string())
                i += 1

        return " ".join(result)

    def _names_assigned_to(self, node: nodes.NodeNG) -> List[str]:
        return sorted([get_name(t) for t in get_assigned_to(node)])

    def _get_then_orelse(self, node: nodes.If) -> Optional[Tuple[nodes.NodeNG, nodes.NodeNG]]:
        if len(node.body) != 1 or len(node.orelse) > 1:
            return None

        then = node.body[0]

        if len(node.orelse) == 1:
            return then, node.orelse[0]

        after_if = node.next_sibling()
        if not isinstance(then, nodes.Return) or not isinstance(after_if, nodes.Return):
            return None

        return then, after_if

    def _refactored_cond_from_then_orelse(
        self, node: nodes.If, then_value: nodes.NodeNG, orelse_value: nodes.NodeNG
    ) -> Optional[Tuple[str, bool]]:
        then_bool_value = then_value.value if self._is_bool(then_value) else None
        orelse_bool_value = orelse_value.value if self._is_bool(orelse_value) else None

        if then_bool_value is None and orelse_bool_value is None:
            return None

        if then_bool_value is not None and orelse_bool_value is not None:
            if then_bool_value and not orelse_bool_value:
                return self._get_refactored(node.test), True
            if not then_bool_value and orelse_bool_value:
                return self._get_refactored("not", node.test), True
            return None

        if then_bool_value is not None and then_bool_value:
            return self._get_refactored(node.test, "or", orelse_value), False

        if then_bool_value is not None and not then_bool_value:
            return self._get_refactored("not", node.test, "and", orelse_value), False

        if orelse_bool_value is not None and orelse_bool_value:
            return self._get_refactored("not", node.test, "or", then_value), False

        if orelse_bool_value is not None and not orelse_bool_value:
            return self._get_refactored(node.test, "and", then_value), False

        assert False, "unreachable"

    def _merge_nested(self, node: nodes.If) -> None:
        refactored = self._get_refactored(node.test, "and", node.body[0].test)
        self.add_message("simplifiable-if-nested", node=node, args=refactored)

    def _same_values(self, node1: nodes.NodeNG, node2: nodes.NodeNG) -> bool:
        return (node1 is None and node2 is None) or (
            node1 is not None and node2 is not None and node1.as_string() == node2.as_string()
        )

    def _mergeable(self, node1: nodes.NodeNG, node2: nodes.NodeNG) -> bool:
        if not isinstance(node1, type(node2)):
            return False

        if isinstance(node1, nodes.Return):
            return node1.value is not None and node2.value is not None

        if is_any_assign(node1) and not isinstance(node1, nodes.For):
            assigned_to1 = get_assigned_to(node1)
            assigned_to2 = get_assigned_to(node2)
            return (
                len(assigned_to1) == 1
                and len(assigned_to1) == len(assigned_to2)
                and isinstance(assigned_to1[0], type(assigned_to2[0]))
                and not isinstance(assigned_to1[0], nodes.Subscript)
                and self._names_assigned_to(node1) == self._names_assigned_to(node2)
            )

        return False

    def _merge_sequential(self, node: nodes.If) -> None:
        second_if = node.next_sibling()

        if (
            isinstance(node.test, nodes.BoolOp)
            or isinstance(second_if.test, nodes.BoolOp)
            or len(second_if.orelse) != 0
            or len(node.body) != 1
            or len(second_if.body) != 1
        ):
            return

        body1 = node.body[0]
        body2 = second_if.body[0]
        if (
            not isinstance(body1, type(body2))
            or not isinstance(body1, nodes.Return)
            or not self._same_values(body1.value, body2.value)
        ):
            return

        parent_body = node.parent.body
        if (
            all(
                isinstance(n, nodes.If)
                or isinstance(n, nodes.Return)
                or (is_any_assign(n) and not isinstance(n, nodes.For))
                for n in parent_body
            )
            and sum(1 if isinstance(n, nodes.If) else 0 for n in parent_body) > 2
        ):
            return

        refactored = self._get_refactored(node.test, "or", second_if.test)
        self.add_message("simplifiable-if-seq", node=node, args=(refactored))

    def _is_just_returning_if(self, node: Optional[nodes.NodeNG]) -> bool:
        return (
            node is not None
            and isinstance(node, nodes.If)
            and isinstance(node.body[-1], nodes.Return)
        )

    @only_required_for_messages(
        "simplifiable-if-return",
        "simplifiable-if-return-conj",
        "simplifiable-if-assignment",
        "simplifiable-if-assignment-conj",
        "simplifiable-if-pass",
        "no-value-in-one-branch-return",
        "simplifiable-if-nested",
        "simplifiable-if-seq",
    )
    def visit_if(self, node: nodes.If) -> None:
        if len(node.orelse) == 0:
            if (
                len(node.body) == 1
                and isinstance(node.body[0], nodes.If)
                and len(node.body[0].orelse) == 0
            ):
                self._merge_nested(node)
                return

            if node.next_sibling() is not None and isinstance(node.next_sibling(), nodes.If):
                self._merge_sequential(node)
                return

        if len(node.body) == 1 and isinstance(node.body[0], nodes.Pass) and len(node.orelse) > 0:
            self.add_message(
                "simplifiable-if-pass", node=node, args=(self._get_refactored("not", node.test))
            )
            return

        then_orelse = self._get_then_orelse(node)

        if not then_orelse:
            return

        then, orelse = then_orelse

        if (
            isinstance(then, nodes.Return)
            and isinstance(orelse, nodes.Return)
            and (then.value is None) != (orelse.value is None)
        ):
            self.add_message("no-value-in-one-branch-return", node=node)
            return

        if not self._mergeable(then, orelse):
            return

        if len(node.orelse) == 0 and self._is_just_returning_if(node.previous_sibling()):
            return

        refactored = self._refactored_cond_from_then_orelse(node, then.value, orelse.value)

        if refactored is not None:
            new_cond, only_replaces = refactored
            self._simplifiable_if_message(node, then, new_cond, only_replaces)

    @only_required_for_messages("simplifiable-if-expr", "simplifiable-if-expr-conj")
    def visit_ifexp(self, node: nodes.IfExp) -> None:
        then, orelse = node.body, node.orelse
        assert then is not None and orelse is not None

        if (not isinstance(then, nodes.Const) and not isinstance(then, nodes.Name)) or (
            not isinstance(orelse, nodes.Const) and not isinstance(orelse, nodes.Name)
        ):
            return

        refactored = self._refactored_cond_from_then_orelse(node, then, orelse)
        if refactored is not None:
            new_cond, only_replaces = refactored
            self._simplifiable_if_message(node, then, new_cond, only_replaces)

    SWITCHED_COMPARATOR = {
        ">=": "<=",
        ">": "<",
        "<=": ">=",
        "<": ">",
    }

    EQ_NEQ = {"==", "!="}

    def _get_values_and_comparator(
        self, node: nodes.Compare
    ) -> Tuple[nodes.NodeNG, str, nodes.NodeNG]:
        """
        This method assumes node is a comparison between just two values.
        """
        left, (comp, right) = node.left, node.ops[0]
        return left, comp, right

    def _is_number(self, node: nodes.NodeNG) -> bool:
        const_val = get_const_value(node)
        return isinstance(const_val, (int, float)) and not isinstance(const_val, bool)

    def _get_node_comparator_const_value(self, node: nodes.NodeNG) -> NodeCmpValue:
        """
        Assumes that node is a child of Bool Op.
        If the extraction was not successful returns None.
        It is succesful if and only if it is an inequality between
        a constant and a non-constant.

        When this does not return None, the constant is last.
        """
        if not isinstance(node, nodes.Compare) or len(node.ops) >= 2:
            return None

        left, comp, right = self._get_values_and_comparator(node)

        left_is_number = self._is_number(left)
        right_is_number = self._is_number(right)

        if (
            (left_is_number and right_is_number)
            or (not left_is_number and not right_is_number)
            or comp not in self.SWITCHED_COMPARATOR
        ):
            return None

        if left_is_number:
            left, comp, right = right, self.SWITCHED_COMPARATOR[comp], left

        return left, comp, get_const_value(right)

    def _get_comparator_by_restriction(self, cmp1: str, cmp2: str, more_restrictive: bool) -> str:
        if len(cmp1) > len(cmp2):
            cmp1, cmp2 = cmp2, cmp1

        return cmp1 if more_restrictive else cmp2

    def _remove_redundant_compare(
        self, cmp1: str, const1: Any, cmp2: str, const2: Any, is_and: bool
    ) -> Tuple[str, int]:
        """
        Supposes that cmp1 and cmp2 are comparisons in the same direction (i.e. > and >=, ...)
        and that const1 and const2 are both numric values.

        <is_and> is True when there is <and> operation between the two comparisons and
        is False when there is <or> between them.
        """
        if const1 == const2:
            return [self._get_comparator_by_restriction(cmp1, cmp2, is_and), const1]

        # make the const1 smaller than const2
        if const1 > const2:
            cmp1, const1, cmp2, const2 = cmp2, const2, cmp1, const1

        if cmp1 == ">" or cmp1 == ">=":
            if is_and:
                return [cmp2, const2]
            return [cmp1, const1]

        if is_and:
            return [cmp1, const1]

        return [cmp2, const2]

    def _add_abs_if_necessary(self, expr_string: str) -> str:
        if expr_string.startswith("abs("):
            return expr_string + " "

        return "abs(" + expr_string + ") "

    def _check_for_simplification_of_single_or(
        self, const1: Any, cmp1: str, const2: Any, cmp2: str, expr_string: str
    ) -> str:
        """
        Supposes that <expr_string> is a variable name or is of type f(x), where
        f is pure builtin function.

        And checks if an expression <<expr_string> <cmp1> <const1> or
        <expr_string> <cmp2> <const2>> can be simplified.

        Checks for problems like 'f(x) >= 4 or f(x) <= -4' and suggests to
        replace them with 'abs(f(x)) >= 4'.

        Returns the suggestion as a string.
        """
        new_expr = "True"

        if cmp1 in {">=", ">"}:
            if const1 > 0:
                new_expr = self._add_abs_if_necessary(expr_string) + cmp1 + " " + str(const1)
        else:
            if const1 < 0:
                new_expr = self._add_abs_if_necessary(expr_string) + cmp2 + " " + str(const2)

        return new_expr

    def _check_for_simplification_of_single_and(
        self, const1: Any, cmp1: str, const2: Any, cmp2: str, expr_string: str
    ) -> str:
        """
        Supposes that <expr_string> is a variable name or is of type f(x), where
        f is pure builtin function. And also supposes that cmp1 and cmp2 are in
        opposite directions.

        And checks if an expression <<expr_string> <cmp1> <const1> and
        <expr_string> <cmp2> <const2>> can be simplified.

        Checks for problems like 'f(x) < 4 and f(x) > -4' and suggests to
        replace them with 'abs(f(x)) < 4'.

        Returns the suggestion as a string.
        """
        new_expr = "False"

        if cmp1 in {">=", ">"}:
            if const1 < 0:
                new_expr = self._add_abs_if_necessary(expr_string) + cmp2 + " " + str(const2)
        else:
            if const1 > 0:
                new_expr = self._add_abs_if_necessary(expr_string) + cmp1 + " " + str(const1)

        return new_expr

    def _check_if_always_true_or_false(
        self, const1: Any, cmp1: str, const2: Any, cmp2: str, expr_string: str, op: str
    ) -> str:
        if const1 == const2:
            if len(cmp1) == 1 and len(cmp2) == 1 and op == "or":
                return expr_string + " != " + str(const1)

            if len(cmp1) == 2 and len(cmp2) == 2 and op == "and":
                return expr_string + " == " + str(const1)

            if op == "or":
                return "True"

            return "False"

        if op == "and" and (
            (const1 < const2 and cmp1[0] == "<") or (const1 > const2 and cmp2[0] == "<")
        ):
            return "False"

        if op == "or" and (
            (const1 < const2 and cmp1[0] == ">") or (const1 > const2 and cmp2[0] == ">")
        ):
            return "True"

        return ""

    def _get_group_as_string(
        self, expr_string: str, group: List[Tuple[Comparison, Value]], op: str
    ) -> str:
        result = []

        for i in range(len(group)):
            cmp, val = group[i]
            result.extend([expr_string, cmp, str(val)])
            if i < len(group) - 1:
                result.append(op)

        return " ".join(result)

    def _suggestion_for_group_of_boolops(
        self, expr_string: str, group: List[Tuple[Comparison, Value]], node: nodes.BoolOp
    ) -> None:
        # We'll use commutativity of boolean operations and group the comparisons in same direction
        # (< with <= and > with >=) and simplify the grouped parts and then we can simplify the rest.
        cmp_less, val1 = None, None
        cmp_greater, val2 = None, None
        changed = False

        for cmp, val in group:
            if cmp[0] == "<":
                if cmp_less is None:
                    cmp_less, val1 = cmp, val
                    continue
                changed = True
                cmp_less, val1 = self._remove_redundant_compare(
                    cmp_less, val1, cmp, val, node.op == "and"
                )
            else:
                if cmp_greater is None:
                    cmp_greater, val2 = cmp, val
                    continue
                changed = True
                cmp_greater, val2 = self._remove_redundant_compare(
                    cmp_greater, val2, cmp, val, node.op == "and"
                )

        group_string = self._get_group_as_string(expr_string, group, node.op)

        if cmp_less is None:
            self.add_message(
                "redundant-compare-in-condition",
                node=node,
                args=(group_string, " ".join([expr_string, cmp_greater, str(val2)])),
            )
            return

        if cmp_greater is None:
            self.add_message(
                "redundant-compare-in-condition",
                node=node,
                args=(group_string, " ".join([expr_string, cmp_less, str(val1)])),
            )
            return

        new_expr = self._check_if_always_true_or_false(
            val1, cmp_less, val2, cmp_greater, expr_string, node.op
        )

        if len(new_expr) != 0:
            self.add_message(
                (
                    "redundant-compare-in-condition"
                    if new_expr == "True" or new_expr == "False"
                    else "using-compare-instead-of-equal"
                ),
                node=node,
                args=(group_string, new_expr),
            )
            return

        if val1 != -val2 or cmp_less != self.SWITCHED_COMPARATOR[cmp_greater]:
            if changed:
                self.add_message(
                    "redundant-compare-in-condition",
                    node=node,
                    args=(
                        group_string,
                        " ".join(
                            [
                                expr_string,
                                cmp_less,
                                str(val1),
                                node.op,
                                expr_string,
                                cmp_greater,
                                str(val2),
                            ]
                        ),
                    ),
                )
            return

        if node.op == "and":
            new_expr = self._check_for_simplification_of_single_and(
                val1, cmp_less, val2, cmp_greater, expr_string
            )
        else:
            new_expr = self._check_for_simplification_of_single_or(
                val1, cmp_less, val2, cmp_greater, expr_string
            )

        self.add_message(
            "simplifiable-with-abs",
            node=node,
            args=(group_string, new_expr),
        )

    def _is_pure_node(self, node: nodes.NodeNG):
        """
        Note: This method does not check children of the node, just the node itself.
        """
        return (
            isinstance(node, nodes.Name)
            or isinstance(node, nodes.Const)
            or isinstance(node, nodes.BinOp)
            or isinstance(node, nodes.BoolOp)
            or isinstance(node, nodes.UnaryOp)
            or isinstance(node, nodes.Compare)
            or (isinstance(node, nodes.Subscript) and node.ctx == Context.Load)
            or isinstance(node, nodes.Attribute)
            or (
                isinstance(node, nodes.Call)
                and len(node.keywords) == 0
                and len(node.kwargs) == 0
                and len(node.starargs) == 0
                and is_pure_builtin(node.func)
            )
        )

    def _is_pure_expression(self, node: nodes.NodeNG) -> bool:
        if not self._is_pure_node(node):
            return False

        children = node.get_children()
        if isinstance(node, nodes.Call):
            # the first child of Call is always the function itself and we know it is pure by now.
            next(children)

        for child in children:
            if not self._is_pure_expression(child):
                return False

        return True

    def _group_and_check_by_representation(
        self, comparison_operands: List[NodeCmpValue], node: nodes.BoolOp
    ) -> None:
        already_checked: Set[int] = set()
        current_group: List[Tuple[Comparison, Value]] = []

        for i in range(len(comparison_operands)):
            if i in already_checked:
                continue

            expr1, cmp1, const1 = comparison_operands[i]
            current_group = [(cmp1, const1)]
            for j in range(i + 1, len(comparison_operands)):
                if j in already_checked:
                    continue

                expr2, cmp2, const2 = comparison_operands[j]
                # I could have done something more, but thanks to how as_string returns same string
                # even if there are some additional parentheses or spaces this seems good enough.
                # Could be improved with checking the actual structure and types of expr1 and expr2,
                # but with the knowledge that all the operands are pure and are from the same bool
                # expression we can safely just check using as_string().
                if expr1.as_string() == expr2.as_string():
                    current_group.append((cmp2, const2))
                    already_checked.add(j)

            if len(current_group) < 2:
                continue

            self._suggestion_for_group_of_boolops(expr1.as_string(), current_group, node)

    def _update_comparisons(
        self,
        comparisons: SameDirComparisons,
        expr1: str,
        comparator: str,
        expr2: str,
        comparison_switched: bool,
    ):
        if expr1 in comparisons:
            if comparator in comparisons[expr1]:
                comparisons[expr1][comparator][0].append(expr2)
                comparisons[expr1][comparator][1].append(comparison_switched)
            else:
                comparisons[expr1][comparator] = ([expr2], [comparison_switched])
        else:
            comparisons[expr1] = {comparator: ([expr2], [comparison_switched])}

    def _switch_if_necessary(
        self, expr1: str, cmp: str, expr2: str, was_switched: bool
    ) -> List[str]:
        if was_switched:
            return [expr2, self.SWITCHED_COMPARATOR[cmp], expr1]

        return [expr1, cmp, expr2]

    def _get_original_group_as_string(
        self,
        expr_string: str,
        cmp: str,
        group: List[ExprRepresentation],
        was_switched: List[bool],
        compared_with_num: Optional[Tuple[str, bool]],
        op: str,
    ) -> str:
        result = []

        for i in range(len(group)):
            result.extend(self._switch_if_necessary(expr_string, cmp, group[i], was_switched[i]))

            if i < len(group) - 1:
                result.append(op)

        if compared_with_num is not None:
            result.append(op)
            result.extend(
                self._switch_if_necessary(
                    expr_string,
                    cmp,
                    compared_with_num[0],
                    compared_with_num[1],
                )
            )

        return " ".join(result)

    def _put_into_max_or_min(
        self,
        function: str,
        expression: str,
        comparison: str,
        expressions: List[str],
        compared_with_num: Optional[Tuple[str, bool]],
        was_switched: List[bool],
    ) -> str:
        """
        the <function> should contain either "max(" or "min("
        """
        result: List[str] = [function]
        all_switched_or_all_not_switched = True

        for i in range(len(expressions)):
            if expressions[i].startswith(function):
                result.append(expressions[i][4:-1])
            else:
                result.append(expressions[i])

            all_switched_or_all_not_switched &= was_switched[i] == was_switched[0]

            if i < len(expressions) - 1:
                result.append(", ")

        if compared_with_num is not None:
            all_switched_or_all_not_switched &= compared_with_num[1] == was_switched[0]

            result.append(", ")
            result.append(compared_with_num[0])

        result.append(")")

        if (all_switched_or_all_not_switched and was_switched[0]) or (
            not all_switched_or_all_not_switched and comparison[0] == ">"
        ):
            return "".join(result) + f" {self.SWITCHED_COMPARATOR[comparison]} {expression}"

        return f"{expression} {comparison} " + "".join(result)

    def _make_suggestion_for_using_max_min_if_possible(
        self,
        comparisons_with_numbers: SameDirComparisons,
        comparisons: SameDirComparisons,
        node: nodes.BoolOp,
    ) -> None:
        """
        Makes suggestion only for the comparisons that will maximize the number of
        comparisons put together with max (ie. in here: "x > a and b < x and b < 1
        and a > b" would suggest "x > a and b < min(x, 1, a)" instead of
        "x > max(a, b) and b < 1 and b < a"). And makes only one suggestion (after
        finding the suggestion as described earlier the method just returns).
        """
        expression: Optional[str] = None
        comparison: Optional[str] = None
        expressions: Optional[List[str]] = None
        was_switched: Optional[List[bool]] = None
        compared_with_num: Optional[Tuple[str, bool]] = None

        for expr, compared_with in comparisons.items():
            for cmp, exprs in compared_with.items():
                if len(exprs[0]) < 2:
                    continue

                comparison_count = len(exprs[0])
                compared_with_1_number = False
                # note: when it's compared with multiple numbers the redundant compare in condition
                # or simplifiable-with-abs could make a suggestion and I don't want to make
                # conflicting suggestions.

                if (
                    expr in comparisons_with_numbers
                    and len(comparisons_with_numbers[expr]) == 1
                    and cmp in comparisons_with_numbers[expr]
                    and len(comparisons_with_numbers[expr][cmp][0]) == 1
                ):
                    comparison_count += 1
                    compared_with_1_number = True

                if expression is None or len(expressions) < comparison_count:
                    expression = expr
                    comparison = cmp
                    expressions, was_switched = exprs

                    if compared_with_1_number:
                        compared_with_num = (
                            comparisons_with_numbers[expr][cmp][0][0],
                            comparisons_with_numbers[expr][cmp][1][0],
                        )
                    else:
                        compared_with_num = None

        if expression is None:
            return

        group_string = self._get_original_group_as_string(
            expression, comparison, expressions, was_switched, compared_with_num, node.op
        )

        func = (
            "max("
            if (
                (node.op == "and" and comparison[0] == ">")
                or (node.op == "or" and comparison[0] == "<")
            )
            else "min("
        )

        self.add_message(
            "redundant-compare-avoidable-with-max-min",
            node=node,
            args=(
                group_string,
                self._put_into_max_or_min(
                    func, expression, comparison, expressions, compared_with_num, was_switched
                ),
            ),
        )

    def _is_equality_or_inequality(self, node: nodes.NodeNG) -> bool:
        return (
            isinstance(node, (nodes.Compare))
            and len(node.ops) == 1
            and node.ops[0][0] in self.EQ_NEQ
        )

    def _is_conjunction_of_two_equalities(self, node: nodes.NodeNG) -> bool:
        return (
            isinstance(node, nodes.BoolOp)
            and node.op == "and"
            and len(node.values) == 2
            and self._is_equality_or_inequality(node.values[0])
            and self._is_equality_or_inequality(node.values[1])
        )

    def _add_parentheses_to_operations(self, node: nodes.NodeNG) -> str:
        if isinstance(node, (nodes.BinOp, nodes.BoolOp, nodes.UnaryOp, nodes.Compare)):
            return f"({node.as_string()})"

        return node.as_string()

    def _destructure_mod_and_number_comparison(
        self, node: nodes.Compare
    ) -> Optional[Tuple[ExprRepresentation, int, Comparison, Value]]:
        left, cmp, right = self._get_values_and_comparator(node)

        left_is_number = self._is_number(left)
        right_is_number = self._is_number(right)

        if (left_is_number and right_is_number) or (not left_is_number and not right_is_number):
            return None

        if left_is_number:
            left, cmp, right = right, cmp, left

        if not isinstance(left, nodes.BinOp) or left.op != "%" or not self._is_number(left.right):
            return None

        mod = get_const_value(left.right)
        val = get_const_value(right)

        if not isinstance(mod, int) or not is_modulo_residue(val, mod):
            return None

        if mod == 2 and cmp == "!=":
            cmp = "=="
            val = 1 - val

        return self._add_parentheses_to_operations(left.left), mod, cmp, val

    def _destructure_mod_and_number_comps(
        self, left: nodes.Compare, right: nodes.Compare
    ) -> Optional[
        Tuple[ExprRepresentation, Comparison, Value, ExprRepresentation, Comparison, Value, int]
    ]:
        left_comparison = self._destructure_mod_and_number_comparison(left)
        right_comparison = self._destructure_mod_and_number_comparison(right)

        if (
            left_comparison is None
            or right_comparison is None
            or left_comparison[1] != right_comparison[1]
        ):
            return None

        return (
            left_comparison[0],
            left_comparison[2],
            left_comparison[3],
            right_comparison[0],
            right_comparison[2],
            right_comparison[3],
            left_comparison[1],
        )

    def _insert_conjuncted_mod_comparison(
        self,
        modulo_connected_with_and: List[PairsOfModComparisonsWithNum],
        expr1: ExprRepresentation,
        cmp1: Comparison,
        val1: Value,
        expr2: ExprRepresentation,
        cmp2: Comparison,
        val2: Value,
        mod: int,
        comparison_pair_string: str,
    ) -> None:
        for pair in modulo_connected_with_and:
            if pair.modulo != mod:
                continue

            if pair.expression_pair == (expr1, expr2):
                pair.update(cmp1, val1, cmp2, val2, comparison_pair_string)
                return

            if pair.expression_pair == (expr2, expr1):
                pair.update(cmp2, val2, cmp1, val1, comparison_pair_string)
                return

        modulo_connected_with_and.append(
            PairsOfModComparisonsWithNum(
                expr1, cmp1, val1, expr2, cmp2, val2, mod, comparison_pair_string
            )
        )

    def _make_suggestion_for_simplifiable_test_by_equals(
        self, modulo_connected_with_and: List[PairsOfModComparisonsWithNum], node: nodes.BoolOp
    ) -> None:
        for pair in modulo_connected_with_and:
            original: Optional[str] = None
            suggestion: Optional[str] = None

            if pair.always_true():
                original = " or ".join([expr for (expr, _) in pair.comparisons_pairs])
                suggestion = "True"
            elif pair.has_part_simplifiable_to_equality():
                original = " or ".join(
                    [
                        expr
                        for (expr, for_comparison) in pair.comparisons_pairs
                        if for_comparison == Comparisons.EQUALITY
                    ]
                )
                suggestion = f"{pair.expression_pair[0]} % {pair.modulo} == {pair.expression_pair[1]} % {pair.modulo}"
            elif pair.has_part_simplifiable_to_inequality():
                original = " or ".join(
                    [
                        expr
                        for (expr, for_comparison) in pair.comparisons_pairs
                        if for_comparison == Comparisons.INEQUALITY
                    ]
                )
                suggestion = f"{pair.expression_pair[0]} % {pair.modulo} != {pair.expression_pair[1]} % {pair.modulo}"

            if original is not None:
                self.add_message(
                    "simplifiable-test-by-equals",
                    node=node,
                    args=(original, suggestion),
                )

    def _is_comparison_not_between_numbers(self, node: nodes.NodeNG) -> bool:
        return (
            isinstance(node, nodes.Compare)
            and len(node.ops) == 1
            and node.ops[0][0] in self.SWITCHED_COMPARATOR
            and not self._is_number(node.left)
            and not self._is_number(node.ops[0][1])
        )

    def _check_for_simplification_of_boolop(self, node: nodes.BoolOp) -> None:
        if node.op is None:
            return

        comparison_operands: List[NodeCmpValue] = []
        comparisons_with_numbers: SameDirComparisons = {}
        comparisons: SameDirComparisons = {}

        node_is_or = node.op == "or"
        modulo_connected_with_and: List[PairsOfModComparisonsWithNum] = []

        for value in node.values:
            if not self._is_pure_expression(value):
                return

            operand = self._get_node_comparator_const_value(value)

            if operand is not None:
                comparison_operands.append(operand)
                self._update_comparisons(
                    comparisons_with_numbers,
                    operand[0].as_string(),
                    operand[1],
                    str(operand[2]),
                    value.ops[0][0] != operand[1],
                )
            elif self._is_comparison_not_between_numbers(value):
                self._update_comparisons(
                    comparisons,
                    value.left.as_string(),
                    value.ops[0][0],
                    value.ops[0][1].as_string(),
                    False,
                )
                self._update_comparisons(
                    comparisons,
                    value.ops[0][1].as_string(),
                    self.SWITCHED_COMPARATOR[value.ops[0][0]],
                    value.left.as_string(),
                    True,
                )
            elif node_is_or and self._is_conjunction_of_two_equalities(value):
                tmp = self._destructure_mod_and_number_comps(value.values[0], value.values[1])
                if tmp is not None:
                    self._insert_conjuncted_mod_comparison(
                        modulo_connected_with_and, *tmp, value.as_string()
                    )

        self._group_and_check_by_representation(comparison_operands, node)
        self._make_suggestion_for_using_max_min_if_possible(
            comparisons_with_numbers, comparisons, node
        )
        self._make_suggestion_for_simplifiable_test_by_equals(modulo_connected_with_and, node)

    @only_required_for_messages(
        "simplifiable-with-abs",
        "redundant-compare-in-condition",
        "redundant-compare-avoidable-with-max-min",
        "using-compare-instead-of-equal",
        "simplifiable-test-by-equals",
    )
    def visit_boolop(self, node: nodes.BoolOp) -> None:
        self._check_for_simplification_of_boolop(node)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(SimplifiableIf(linter))



from typing import Any, TypeVar, Generic, List, Iterable, Union, Optional, Tuple, cast, Callable
from functools import reduce
from astroid import nodes, Uninferable  # type: ignore
import sys
import inspect
import operator
from pylint.checkers import utils  # type: ignore

from edulint.linting.analyses.data_dependency import node_to_var

POSTINIT_ARGS = {
    klass: (
        {
            param.name: (
                list
                if param.annotation != inspect.Parameter.empty
                and param.annotation.lower().startswith("list")
                else lambda: None
            )
            for param in inspect.signature(klass.postinit).parameters.values()
            if param.name != "self"
        }
        if klass != nodes.Const and hasattr(klass, "postinit")
        else {}
    )
    for klass in nodes.ALL_NODE_CLASSES
}


def requires_data_dependency_analysis(func):
    def inner(self, node: nodes.NodeNG, *args, **kwargs):
        if not node.root().cfg_loc.var_events.successful:
            return
        func(self, node, *args, **kwargs)

    return inner


def new_node(node_type, **attr_vals):
    attr_vals_before = {
        attr: val for attr, val in attr_vals.items() if attr not in POSTINIT_ARGS[node_type].keys()
    }
    attr_vals_after = {
        attr: attr_vals.get(attr, default()) for attr, default in POSTINIT_ARGS[node_type].items()
    }

    if node_type == nodes.Arguments:
        attr_vals_before["kwarg"] = attr_vals_before.get("kwarg", "")
        attr_vals_before["vararg"] = attr_vals_before.get("vararg", "")
        node = node_type(parent=None, **attr_vals_before)
    elif node_type == nodes.Module:
        node = node_type(**attr_vals_before)
    else:
        if node_type == nodes.ImportFrom:
            attr_vals_before["fromname"] = attr_vals_before.pop("modname")
        node = node_type(
            lineno=0, col_offset=0, end_lineno=0, end_col_offset=0, parent=None, **attr_vals_before
        )

    if not isinstance(node, nodes.Const) and hasattr(node, "postinit"):
        node.postinit(**attr_vals_after)

    return node


EXPRESSION_TYPES = (
    nodes.Subscript,
    nodes.Attribute,
    nodes.BinOp,
    nodes.BoolOp,
    nodes.Call,
    nodes.Compare,
    nodes.Const,
    nodes.List,
    nodes.Dict,
    nodes.Set,
    nodes.GeneratorExp,
    nodes.IfExp,
    nodes.Lambda,
    nodes.Tuple,
    nodes.UnaryOp,
)


def eprint(*args: Any, **kwargs: Any) -> None:
    print(*args, file=sys.stderr, **kwargs)


T = TypeVar("T")


class BaseVisitor(Generic[T]):
    default: T = None  # type: ignore

    def combine(self, results: List[T]) -> T:
        if not results:
            return self.default  # type: ignore
        return results[-1]

    def visit(self, node: nodes.NodeNG) -> T:
        return node.accept(self)  # type: ignore

    def visit_many(self, nodes: Iterable[nodes.NodeNG]) -> T:
        return self.combine([self.visit(node) for node in nodes])


def generic_visit(self: BaseVisitor[T], node: nodes.NodeNG) -> T:
    return self.combine([self.visit(child) for child in node.get_children()])


for name, obj in inspect.getmembers(nodes, inspect.isclass):
    setattr(BaseVisitor, f"visit_{obj.__name__.lower()}", generic_visit)


# solution using inference at
# https://github.com/PyCQA/pylint/blob/ca80f03a43bc39e4cc2c67dc99817b3c9f13b8a6/pylint/checkers/refactoring/recommendation_checker.py
def is_builtin(node: nodes.NodeNG, function: Optional[str] = None) -> bool:
    assert function is None or utils.is_builtin(function)
    node_var = node_to_var(node)
    return node_var is None and (
        function is None or (isinstance(node, nodes.Name) and node.name == function)
    )


PURE_FUNCTIONS = {
    "abs",
    "max",
    "min",
    "round",
    "sqrt",
    "len",
    "sum",
    "sorted",
    "reversed",
    "int",
    "float",
    "str",
    "ord",
    "chr",
}

EXPR_FUNCTIONS = PURE_FUNCTIONS | {"all", "any", "map", "filter"}


def is_pure_builtin(node: nodes.NodeNG) -> bool:
    inferred = utils.safe_infer(node)
    return is_builtin(node) and inferred and inferred.name in PURE_FUNCTIONS


def is_multi_assign(node: nodes.NodeNG) -> bool:
    return hasattr(node, "targets")


def is_assign(node: nodes.NodeNG) -> bool:
    return hasattr(node, "target")


def is_any_assign(node: nodes.NodeNG) -> bool:
    return is_assign(node) or is_multi_assign(node)


def get_assigned_to(node: nodes.NodeNG) -> List[nodes.NodeNG]:
    if is_multi_assign(node):
        return cast(List[nodes.NodeNG], node.targets)
    if is_assign(node):
        return [node.target]
    return []


Named = Union[nodes.Name, nodes.Attribute, nodes.AssignName]


def is_named(node: nodes.NodeNG) -> bool:
    return hasattr(node, "name") or hasattr(node, "attrname")


def get_name(node: Named) -> str:
    return node.as_string()


def get_range_params(
    node: nodes.NodeNG,
) -> Optional[Tuple[nodes.NodeNG, nodes.NodeNG, nodes.NodeNG]]:
    if (
        not isinstance(node, nodes.Call)
        or node.func.as_string() != "range"
        or len(node.args) < 1
        or len(node.args) > 3
    ):
        return None

    default_start = new_node(nodes.Const, value=0)
    default_step = new_node(nodes.Const, value=1)

    if len(node.args) == 1:
        return default_start, node.args[0], default_step

    if len(node.args) == 2:
        return node.args[0], node.args[1], default_step

    if len(node.args) == 3:
        return node.args[0], node.args[1], node.args[2]

    assert False, "unreachable"


# based on https://docs.python.org/3/library/operator.html
UNARY_SYMBOL_TO_OP = {
    "+": operator.pos,
    "-": operator.neg,
    "not": operator.not_,
    "~": operator.invert,
}

BINARY_SYMBOL_TO_OP = {
    "+": operator.add,
    "/": operator.truediv,
    "//": operator.floordiv,
    "&": operator.and_,
    "^": operator.xor,
    "|": operator.or_,
    "**": operator.pow,
    "is": operator.is_,
    "is not": operator.is_not,
    "<<": operator.lshift,
    "%": operator.mod,
    "*": operator.mul,
    ">>": operator.rshift,
    "-": operator.sub,
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
    "and": lambda lt, rt: lt and rt,
    "or": lambda lt, rt: lt or rt,
}


def get_const_value_rec(node: Any) -> Any:
    if not isinstance(node, nodes.NodeNG):
        return node

    if isinstance(node, nodes.Const):
        return node.value

    if isinstance(node, nodes.UnaryOp):
        return UNARY_SYMBOL_TO_OP[node.op](get_const_value_rec(node.operand))

    if isinstance(node, nodes.BinOp):
        return BINARY_SYMBOL_TO_OP[node.op](
            get_const_value_rec(node.left), get_const_value_rec(node.right)
        )

    if isinstance(node, nodes.BoolOp):
        assert len(node.values) > 1
        return reduce(BINARY_SYMBOL_TO_OP[node.op], map(get_const_value_rec, node.values))

    raise ValueError(f"{type(node)} cannot be evaluated")


def get_const_value(node: Any) -> Any:
    try:
        return get_const_value_rec(node)
    except ValueError:
        return None


def infer_to_value(node: nodes.NodeNG) -> Optional[nodes.NodeNG]:
    if isinstance(node, nodes.Name):
        inferred = utils.safe_infer(node)
        return None if inferred is Uninferable else inferred

    if isinstance(
        node,
        (
            nodes.Const,
            nodes.List,
            nodes.Set,
            nodes.Dict,
            nodes.ListComp,
            nodes.DictComp,
            nodes.Call,
        ),
    ):
        return node

    return None


def is_parents_elif(node: nodes.If) -> bool:
    parent = node.parent
    return isinstance(parent, nodes.If) and parent.has_elif_block() and parent.orelse[0] == node


def has_else_block(node: Union[nodes.For, nodes.While, nodes.If, nodes.IfExp]):
    if isinstance(node, nodes.IfExp):
        return True
    return len(node.orelse) > 0 and (not isinstance(node, nodes.If) or not node.has_elif_block())


def get_lines_between(first: nodes.NodeNG, last: nodes.NodeNG, including_last: bool) -> int:
    assert first.fromlineno <= last.fromlineno

    if including_last:
        return last.tolineno - first.fromlineno + 1
    return last.fromlineno - first.fromlineno


def is_main_block(statement: nodes.If) -> bool:
    """
    Return whether or not <statement> is the main block.
    """
    return (
        isinstance(statement, nodes.If)
        and isinstance(statement.test, nodes.Compare)
        and isinstance(statement.test.left, nodes.Name)
        and isinstance(statement.test.left, nodes.Name)
        and statement.test.left.name == "__name__"
        and len(statement.test.ops) == 1
        and statement.test.ops[0][0] == "=="
        and isinstance(statement.test.ops[0][1], nodes.Const)
        and statement.test.ops[0][1].value == "__main__"
    )


def is_block_comment(stmt: nodes.NodeNG) -> bool:
    return (
        isinstance(stmt, nodes.Expr)
        and isinstance(stmt.value, nodes.Const)
        and isinstance(stmt.value.value, str)
    )


def get_statements_count(
    node: Union[nodes.NodeNG, List[nodes.NodeNG]], include_defs: bool, include_name_main: bool
) -> int:
    def count(nodes: List[nodes.NodeNG]) -> int:
        return sum(get_statements_count(node, include_defs, include_name_main) for node in nodes)

    if isinstance(node, list):
        return count(node)

    if isinstance(node, (nodes.ClassDef, nodes.FunctionDef)):
        return 1 + count(node.body) if include_defs else 0

    if isinstance(node, (nodes.Import, nodes.ImportFrom)):
        return 1 if include_defs else 0

    if isinstance(node, (nodes.For, nodes.While, nodes.If)):
        if is_main_block(node) and not include_name_main:
            return 0
        return 1 + count(node.body) + (1 if has_else_block(node) else 0) + count(node.orelse)

    if isinstance(node, nodes.Module):
        return count(node.body)

    if isinstance(node, (nodes.Try, nodes.TryStar)):
        return (
            2
            + count(node.body)
            + sum(count(h.body) for h in node.handlers)
            + count(node.orelse)
            + count(node.finalbody)
        )

    if isinstance(node, nodes.With):
        return 1 + count(node.body)

    return 1


class TokenCountingVisitor(BaseVisitor[int]):
    default = 0

    @classmethod
    def combine(cls, results: List[int]) -> int:
        return sum(results) + 1

    def _visit_with_else(self, node: Union[nodes.If, nodes.For, nodes.While]) -> int:
        return self.visit_many(node.get_children()) + (1 if has_else_block(node) else 0)

    def visit_if(self, node: nodes.If) -> int:
        return self._visit_with_else(node)

    def visit_for(self, node: nodes.For) -> int:
        return self._visit_with_else(node)

    def visit_while(self, node: nodes.While) -> int:
        return self._visit_with_else(node)

    def visit_ifexp(self, node: nodes.IfExp) -> int:
        return self._visit_with_else(node)

    def visit_expr(self, node: nodes.Expr) -> int:
        return self.visit(node.value)


def get_token_count(node: Union[nodes.NodeNG, List[nodes.NodeNG]]) -> int:
    visitor = TokenCountingVisitor()
    if isinstance(node, (list, tuple)):
        return visitor.visit_many(node) - 1
    else:
        return visitor.visit(node)


NEGATED_OP = {
    ">=": "<",
    "<=": ">",
    ">": "<=",
    "<": ">=",
    "==": "!=",
    "!=": "==",
    "is": "is not",
    "is not": "is",
    "in": "not in",
    "not in": "in",
    "and": "or",
    "or": "and",
}


def is_negation(lt: nodes.NodeNG, rt: nodes.NodeNG, negated_rt: bool) -> bool:
    def ops_match(lt: nodes.NodeNG, rt: nodes.NodeNG, lt_transform: Callable[[str], str]) -> bool:
        return all(lt_transform(lt_op) == rt_op for (lt_op, _), (rt_op, _) in zip(lt.ops, rt.ops))

    def to_values(node: nodes.NodeNG) -> List[nodes.NodeNG]:
        return [node.left] + [val for _, val in node.ops]

    def all_are_negations(
        lt_values: List[nodes.NodeNG], rt_values: List[nodes.NodeNG], new_rt_negated: bool
    ) -> bool:
        return all(is_negation(ll, rr, new_rt_negated) for ll, rr in zip(lt_values, rt_values))

    def strip_nots(node: nodes.NodeNG, negated_rt: bool) -> Tuple[nodes.NodeNG, bool]:
        while isinstance(node, nodes.UnaryOp) and node.op == "not":
            negated_rt = not negated_rt
            node = node.operand
        return node, negated_rt

    def is_mod_two(node: nodes.NodeNG):
        return isinstance(node, nodes.BinOp) and node.op == "%" and get_const_value(node.right) == 2

    # TODO allow x % 2 > 0
    def is_eq_mod_two(node: nodes.NodeNG):
        if len(node.ops) != 1:
            return False

        left, (op, right) = node.left, node.ops[0]
        if op not in ("==", "!=") or is_mod_two(left) == is_mod_two(right):
            return False

        # MAYBE check if value equal to mod is 0 or 1?
        return True

    def mod_two_negation(lt: nodes.NodeNG, rt: nodes.NodeNG):
        if not is_eq_mod_two(lt) or not is_eq_mod_two(rt):
            return False

        lt_lt, (_, lt_rt) = lt.left, lt.ops[0]
        rt_lt, (_, rt_rt) = rt.left, rt.ops[0]

        lt_mod, lt_val = (lt_lt, lt_rt) if is_mod_two(lt_lt) else (lt_rt, lt_lt)
        rt_mod, rt_val = (rt_lt, rt_rt) if is_mod_two(rt_lt) else (rt_rt, rt_lt)

        return are_identical(lt_mod.left, rt_mod.left) and not are_identical(lt_val, rt_val)

    lt, negated_rt = strip_nots(lt, negated_rt)
    rt, negated_rt = strip_nots(rt, negated_rt)

    if not isinstance(lt, type(rt)):
        return False

    if isinstance(lt, nodes.BoolOp):
        if len(lt.values) == len(rt.values) and (
            (negated_rt and lt.op == rt.op) or (not negated_rt and NEGATED_OP[lt.op] == rt.op)
        ):
            return all_are_negations(lt.values, rt.values, negated_rt)
        return False

    if isinstance(lt, nodes.Compare):
        if len(lt.ops) != len(rt.ops):
            return False

        if negated_rt and ops_match(lt, rt, lambda op: op):
            return all_are_negations(to_values(lt), to_values(rt), negated_rt)

        if not negated_rt and ops_match(lt, rt, lambda op: NEGATED_OP[op]):
            return all_are_negations(to_values(lt), to_values(rt), not negated_rt)

        if not negated_rt and mod_two_negation(lt, rt):
            return True

        return False

    return negated_rt and are_identical(lt, rt)


def are_identical(
    block1: Union[nodes.NodeNG, List[nodes.NodeNG]], block2: Union[nodes.NodeNG, List[nodes.NodeNG]]
) -> bool:
    if isinstance(block1, nodes.NodeNG):
        assert isinstance(block2, nodes.NodeNG)
        block1 = [block1]
        block2 = [block2]

    strings1 = [n.as_string() for n in block1]
    strings2 = [n.as_string() for n in block2]

    return strings1 == strings2



from typing import Dict
from pathlib import Path
import os

import tomli
import tomli_w

import thonny_messages


SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))
FILENAME_TOML = os.path.join(SCRIPT_PATH, "thonny_pylint.toml")
EDULINT_TOML = os.path.join(SCRIPT_PATH, "../edulint_thonny.toml")


def extract_msgs_from_thonny(reload_from_source: bool = False) -> Dict[str, Dict[str, str]]:
    if not Path(FILENAME_TOML).is_file():
        reload_from_source = True

    if reload_from_source:
        checks = thonny_messages.checks_by_id
        for check in checks.values():
            if check.get("tho_xpln"):
                check["customized_by_thonny"] = True

        with open(FILENAME_TOML, "wb") as f:
            tomli_w.dump(checks, f, multiline_strings=True)

    with open(FILENAME_TOML, "rb") as f:
        checks = tomli.load(f)

    return checks


def convert_to_edulint():
    with open(FILENAME_TOML, "rb") as f:
        checks = tomli.load(f)

    checks = dict(filter(lambda x: x[1].get("customized_by_thonny"), checks.items()))

    answer = {}
    for key, data in checks.items():
        answer[key] = {
            "why": data["tho_xpln"],
            "examples": "",
        }

    with open(EDULINT_TOML, "wb") as f:
        tomli_w.dump(answer, f, multiline_strings=True)


def process_from_stored_data():
    extract_msgs_from_thonny(reload_from_source=True)
    convert_to_edulint()


if __name__ == "__main__":
    process_from_stored_data()



from .linters import Linter
from .config.config import ImmutableConfig, get_config_one, get_config_many
from .linting.problem import Problem
from .linting.linting import lint_one, lint_many
from .explanations import get_explanations

__all__ = [
    "Linter",
    "ImmutableConfig",
    "get_config_one",
    "get_config_many",
    "Problem",
    "lint_one",
    "lint_many",
    "get_explanations",
]

__version__ = "4.1.5"
__version_info__ = tuple(map(int, __version__.split(".")))



from typing import TYPE_CHECKING, List, Tuple, Dict, Optional, Iterator, Set

from astroid import nodes  # type: ignore
from pylint.checkers import BaseChecker  # type: ignore

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.analyses.antiunify import antiunify, cprint  # noqa: F401
from edulint.linting.analyses.cfg.utils import (
    get_stmt_locs,
    syntactic_children_locs,
    successors_from_loc,
)
from edulint.linting.checkers.utils import is_block_comment, requires_data_dependency_analysis
from edulint.linting.checkers.duplication.duplicate_if import duplicate_in_if
from edulint.linting.checkers.duplication.duplicate_sequence import similar_to_loop
from edulint.linting.checkers.duplication.duplicate_block import similar_to_block
from edulint.linting.checkers.duplication.utils import (
    is_duplication_candidate,
    get_loop_repetitions,
)


### helpers


def candidate_fst(nodes: List[nodes.NodeNG]) -> Iterator[nodes.NodeNG]:
    yield from enumerate(nodes)


def candidate_snd(nodes: List[nodes.NodeNG], i: int) -> Iterator[Tuple[int, nodes.NodeNG]]:
    fst = nodes[i]
    j = i + 1

    while j < len(nodes) and fst.tolineno >= nodes[j].fromlineno:
        j += 1

    for j in range(j, len(nodes)):
        yield j, nodes[j]


def get_siblings(node: nodes.NodeNG) -> List[nodes.NodeNG]:
    siblings = []
    sibling = node
    while sibling is not None:
        if break_on_stmt(sibling):
            break
        if not skip_stmt(sibling):
            siblings.append(sibling)
        sibling = sibling.next_sibling()

    assert len(siblings) > 0
    return siblings


def get_memoized_siblings(
    siblings: Dict[nodes.NodeNG, List[nodes.NodeNG]], node: nodes.NodeNG
) -> List[nodes.NodeNG]:
    sibs = siblings.get(node)
    if sibs is not None:
        return sibs

    sibs = siblings.get(node.previous_sibling())
    if sibs is not None:
        sibs = sibs[1:]
        siblings[node] = sibs
        return sibs

    sibs = get_siblings(node)
    siblings[node] = sibs
    return sibs


def get_stmt_range(
    stmt_to_index: Dict[nodes.NodeNG, int], nodes: List[nodes.NodeNG]
) -> Optional[Tuple[int, int]]:
    last_i = stmt_to_index.get(nodes[-1])
    return stmt_to_index[nodes[0]], last_i + 1 if last_i is not None else None


def overlap(
    stmt_nodes: List[nodes.NodeNG], range1: Tuple[int, int], range2: Tuple[int, int]
) -> bool:
    first1, last1 = range1
    first2, last2 = range2
    if last1 is None or last2 is None:
        return True

    last_node = stmt_nodes[last1 - 1]
    first_node = stmt_nodes[first2]
    return last_node.tolineno >= first_node.fromlineno


def break_on_stmt(node: nodes.NodeNG) -> bool:
    return isinstance(node, (nodes.Assert, nodes.ClassDef))


def skip_stmt(node: nodes.NodeNG) -> bool:
    return (
        is_block_comment(node)
        or isinstance(node, nodes.Pass)
        or (isinstance(node, (nodes.Import, nodes.ImportFrom)) and node.parent == node.root())
        or (
            isinstance(node, (nodes.Assign, nodes.AugAssign, nodes.AnnAssign))
            and sum(len(event.uses) for _var, event in node.cfg_loc.var_events.all()) == 0
        )
    )


def include_in_stmts(node: nodes.NodeNG) -> bool:
    return not break_on_stmt(node) and not skip_stmt(node)


def get_statement_nodes(node: nodes.Module) -> List[nodes.NodeNG]:
    return sorted(
        (
            stmt_loc.node
            for loc in successors_from_loc(
                node.cfg_loc, include_start=True, explore_functions=True, explore_classes=True
            )
            for stmt_loc in get_stmt_locs(loc)
            if stmt_loc is not None and include_in_stmts(stmt_loc.node)
        ),
        key=lambda node: (
            node.fromlineno,
            node.col_offset if node.col_offset is not None else float("inf"),
        ),
    )


def get_duplicate_nodes(ns: List[nodes.NodeNG]) -> Set[nodes.NodeNG]:
    return {loc.node for loc in syntactic_children_locs(ns, include_stmt_locs=True)}


### control functions


def is_similar_to_loop(checker, siblings: List[nodes.NodeNG]) -> bool:
    if (
        (
            not checker.linter.is_message_enabled("similar-block-to-loop-range")
            and not checker.linter.is_message_enabled("similar-block-to-loop-collection")
            and not checker.linter.is_message_enabled("similar-block-to-loop-merge")
        )
        or len(siblings) < 3
        or any(isinstance(node, nodes.FunctionDef) for node in siblings)
    ):
        return False

    for end, to_aunify in get_loop_repetitions(siblings):
        if not is_duplication_candidate(to_aunify):
            continue
        if similar_to_loop(checker, to_aunify):
            return True
    return False


def is_duplicate_in_if(checker, node: nodes.NodeNG) -> Tuple[bool, bool]:
    if not isinstance(node, nodes.If):
        return False, False
    return duplicate_in_if(checker, node)


def get_similar_to_block_candidates(
    checker,
    stmt_nodes: List[nodes.NodeNG],
    stmt_to_index: Dict[nodes.NodeNG, int],
    siblings: Dict[nodes.NodeNG, List[nodes.NodeNG]],
    i: int,
):
    if not checker.linter.is_message_enabled(
        "similar-to-function"
    ) and not checker.linter.is_message_enabled("similar-block-to-call"):
        return

    fst = stmt_nodes[i]
    fst_siblings = get_memoized_siblings(siblings, fst)

    for j, snd in candidate_snd(stmt_nodes, i):
        snd_siblings = get_memoized_siblings(siblings, snd)

        for length in range(min(len(fst_siblings), len(snd_siblings), j - i), 0, -1):
            if length == 1 and (
                isinstance(fst, (nodes.Assign, nodes.Expr))
                or isinstance(snd, (nodes.Assign, nodes.Expr))
            ):
                break

            to_aunify = [tuple(fst_siblings[:length]), tuple(snd_siblings[:length])]
            ranges = [
                get_stmt_range(stmt_to_index, to_aunify[0]),
                get_stmt_range(stmt_to_index, to_aunify[1]),
            ]

            if not overlap(stmt_nodes, ranges[0], ranges[1]) and is_duplication_candidate(
                [stmt_nodes[r1:r2] for r1, r2 in ranges]
            ):
                yield ranges, to_aunify


def is_any_similar_to_block(checker, duplicate: Set[nodes.NodeNG], candidates):
    def may_be_similar_to_loop(ranges):
        return all(last == first for ((_, last), (first, _)) in zip(ranges, ranges[1:]))

    def get_ordered_candidates(candidates):
        filpped_candidates = {}

        for c, id_ in candidates.items():
            if id_ not in filpped_candidates:
                filpped_candidates[id_] = []
            filpped_candidates[id_].append(c)

        for candidate in filpped_candidates.values():
            candidate.sort(key=lambda v: v[0])

        return sorted(
            filpped_candidates.values(),
            key=lambda candidate: (candidate[0][0][0], -candidate[0][0][1]),
        )

    def all_children_of_one_statement(to_aunify: List[List[nodes.NodeNG]]) -> bool:
        last_ancestors = set(to_aunify[0][-1].node_ancestors())
        for parent in to_aunify[0][0].node_ancestors():
            if not isinstance(parent, nodes.Module) and parent in last_ancestors:
                last_same_type_sibling = parent
                while not isinstance(parent, nodes.FunctionDef) and isinstance(
                    last_same_type_sibling.next_sibling(), type(parent)
                ):
                    last_same_type_sibling = last_same_type_sibling.next_sibling()
                from_ = parent.fromlineno
                to_ = last_same_type_sibling.tolineno
                if all(
                    from_ <= sub_aunify[0].fromlineno and sub_aunify[-1].tolineno <= to_
                    for sub_aunify in to_aunify
                ):
                    return True
        return False

    for candidate in get_ordered_candidates(candidates):
        if may_be_similar_to_loop([r for r, _ in candidate]):
            continue

        to_aunify = [list(sub_aunify) for _, sub_aunify in candidate]
        if to_aunify[0][0] in duplicate:
            continue

        if all_children_of_one_statement(to_aunify):
            continue

        if similar_to_block(checker, to_aunify):
            duplicate.update(
                {node for sub_aunify in to_aunify for node in get_duplicate_nodes(sub_aunify)}
            )


class NoDuplicateCode(BaseChecker):  # type: ignore
    name = "no-duplicate-code"
    msgs = {
        "R6501": (
            "The branches of the 'if' statement are identical. Remove the 'if' if it is on purpose, "
            "or change the branches if not.",
            "identical-if-branches",
            "Emitted when all branches of an 'if' statement contain identical code.",
        ),
        "R6502": (
            "Identical code inside all if's branches, move %d lines %s the if.",
            "identical-if-branches-part",
            "Emitted when identical code starts or ends all branches of an if statement.",
        ),
        "R6503": (
            "Identical code inside %d consecutive ifs, join their conditions using 'or'.",
            "identical-seq-ifs",
            "Emitted when several consecutive if statements have identical bodies and thus can be "
            "joined by or in their conditions.",
        ),
        "R6551": (
            "There are %d repetitions of %d similar statements, starting on lines %s. "
            "Extract them to a helper function.",
            "similar-block-to-function",
            "Emitted when similar blocks can be deduplicated using a helper function.",
        ),
        "R6552": (
            "There are %d repetitions of %d similar statements, which can be simplified using a loop. "
            "Consider iterating over '%s'.",
            "similar-block-to-loop-range",
            "Emitted when similar sequential blocks can be deduplicated using a loop.",
        ),
        "R6560": (
            "There are %d repetitions of %d similar statements, which can be simplified using a loop. "
            "Consider iterating over '%s'.",
            "similar-block-to-loop-collection",
            "Emitted when similar sequential blocks can be deduplicated using a loop.",
        ),
        "R6553": (
            "The body of the loop contains duplicate blocks, starting on lines %s. "
            "Avoid the duplication by repeating the outer loop %s times.",
            "similar-block-to-loop-merge",
            "Emitted when the body can be simplified using a loop and "
            "the extra iterations can be performed by iterating the outer loop more times.",
        ),
        "R6554": (
            "Call the existing function '%s' instead of repeating its body.",
            "similar-block-to-call",
            "Emitted when a block can be replaced by a call to an existing function.",
        ),
        "R6555": (
            "Consider using condition directly instead of literals 'True'/'False', "
            "if the if's body has no effect on the condition's value.",
            "similar-if-to-use",
            "Emitted when the two branches of an 'if' statement differ only in True/False, "
            "so the if's condition can be used directly.",
        ),
        "R6556": (
            "The branches of the 'if' statement are similar. Move the 'if' into the branches, "
            "so that it only contains lines %s, but only if the if's body has no effect on the condition's value.",
            "similar-if-into-block",
            "Emitted when an 'if' statement wraps too long part of a block "
            "(similar to identical-if-branches-part, but this one can suggest "
            "moving if into a for loop, for example).",
        ),
        "R6557": (
            "The branches of the 'if' statement are similar. Use condition '%s' "
            "to avoid the duplication.",
            "similar-if-to-untwisted",
            "Emitted when an 'if' statement has an if statement in each branch and "
            "the nested statements have the same branches, only swapped.",
        ),
        "R6558": (
            "The branches of the 'if' statement are similar. Use if expression to decide between "
            "the different values (i.e., \"%s\"), but only if the if's body has no effect on the condition's value.",
            "similar-if-to-expr",
            "Emitted when the duplication between if statement's branches can be avoided "
            "using the if expression.",
        ),
        "R6559": (
            "The branches of the 'if' statement are similar. Use helper variables to store the "
            "values that differ between the branches, or create a helper function.",
            "similar-if-to-extracted",
            "Emitted when the duplication between if statement's branches can be avoided "
            "using helper variables or functions.",
        ),
    }

    @requires_data_dependency_analysis
    def visit_module(self, node: nodes.Module):
        if len(node.body) == 0:
            return

        stmt_nodes = get_statement_nodes(node)
        stmt_to_index = {node: i for i, node in enumerate(stmt_nodes)}

        duplicate = set()
        candidates = {}
        siblings = {}
        for i, fst in candidate_fst(stmt_nodes):
            if fst in duplicate:
                continue

            fst_siblings = get_memoized_siblings(siblings, fst)

            if is_similar_to_loop(self, fst_siblings):
                duplicate.update(get_duplicate_nodes(fst_siblings))
                continue

            any_duplication, inspect_first_if = is_duplicate_in_if(self, fst)
            if any_duplication:
                duplicate.update(
                    get_duplicate_nodes(fst)
                    - (get_duplicate_nodes(fst.body[0]) if inspect_first_if else set())
                )
                continue

            for ranges, to_aunify in get_similar_to_block_candidates(
                self, stmt_nodes, stmt_to_index, siblings, i
            ):
                # TODO or larger?
                id_ = candidates.get((ranges[0], to_aunify[0]), len(candidates))
                candidates[(ranges[0], to_aunify[0])] = id_
                candidates[(ranges[1], to_aunify[1])] = id_

        is_any_similar_to_block(self, duplicate, candidates)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(NoDuplicateCode(linter))



from typing import List, Any, TypeVar, Optional, Union, Tuple
from enum import Enum


class NumberFromZero(Enum):
    def __new__(cls, *args: Any) -> "NumberFromZero":
        value = len(cls.__members__)
        obj = object.__new__(cls)
        obj._value_ = value
        return obj


T = TypeVar("T")
UnionT = Union[bool, List[str], Optional[str], Optional[int]]
ImmutableT = Union[bool, Tuple[str, ...], Optional[str], Optional[int]]


class Option(NumberFromZero):
    CONFIG_FILE = ()
    PYLINT = ()
    FLAKE8 = ()
    ALLOWED_ONECHAR_NAMES = ()
    DISALLOWED_BUILTIN_NAMES = ()
    NO_FLAKE8 = ()
    IGNORE_INFILE_CONFIG_FOR = ()
    EXPORT_GROUPS = ()
    SET_GROUPS = ()
    LANGUAGE_FILE = ()

    def to_name(self) -> str:
        return self.name.lower().replace("_", "-")

    @staticmethod
    def safe_from_name(option_str: str) -> Optional["Option"]:
        aliased = OPTION_ALIASES.get(option_str.lower())
        if aliased is not None:
            return aliased

        for option in Option:
            if option.to_name() == option_str.lower():
                return option
        return None

    @staticmethod
    def from_name(option_str: str) -> "Option":
        option = Option.safe_from_name(option_str)
        if option is not None:
            return option

        assert False, "no such option: " + option_str

    def __int__(self) -> int:
        return self.value  # type: ignore


OPTION_ALIASES = {"config": Option.CONFIG_FILE}

DEFAULT_CONFIG = "default"
BASE_CONFIG = "empty"



from edulint.linting.analyses.cfg.visitor import CFGVisitor
from edulint.linting.analyses.cfg.utils import successors_from_loc
from edulint.linting.analyses.variable_modification import VarModificationAnalysis, VarEventType
from edulint.linting.analyses.data_dependency import name_to_var, modified_in, node_to_var

from typing import List, Tuple, Dict
import astroid  # type: ignore
from astroid import nodes as astns
import pytest


def split_n_lines(nodes: List[astns.NodeNG], lines: int) -> Tuple[List[astns.NodeNG], List[astns.NodeNG]]:
    if lines == 0:
        return [], nodes

    consumed_lines = []
    for i, node in enumerate(nodes):
        consumed_lines.append(node)
        if hasattr(node, "body"):
            consumed_lines_rec, result_rec = split_n_lines(node.body, lines - len(consumed_lines))
            consumed_lines += consumed_lines_rec
        else:
            result_rec = []

        if len(consumed_lines) == lines:
            return consumed_lines, result_rec + nodes[i+1:]
    return consumed_lines, []


@pytest.mark.filterwarnings("ignore:The 'Module.doc' attribute")
@pytest.mark.parametrize("program,init_lines,before_types,after_types", [
    ("x = 5", 1, [astns.Assign], []),
    ("""
x = 5
x += 1""", 1, [astns.Assign], [astns.AugAssign]),
    ("""
x = 5
x = 1""", 1, [astns.Assign], [astns.Assign]),
    ("""
x = 5
y = 1
x += 1""", 2, [astns.Assign, astns.Assign], [astns.AugAssign]),
    ("""
for x in range(10):
    x = 1""", 1, [astns.For], [astns.Assign]),
    ("""
for x in range(10):
    for y in range(11):
        x = 1""", 1, [astns.For], [astns.For]),
    ("""
for x in range(10):
    for y in range(11):
        x = 1""", 2, [astns.For, astns.For], [astns.Assign]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1""", 2, [astns.For, astns.Assign], [astns.For]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1""", 1, [astns.For], [astns.Assign, astns.For]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1
    x += 1""", 1, [astns.For], [astns.Assign, astns.For, astns.AugAssign]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1
    x += 1""", 2, [astns.For, astns.Assign], [astns.For, astns.AugAssign]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1
    x += 1""", 3, [astns.For, astns.Assign, astns.For], [astns.Assign, astns.AugAssign]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1
    x += 1""", 4, [astns.For, astns.Assign, astns.For, astns.Assign], [astns.AugAssign]),
])
def test_split_n_lines(program: str, init_lines: int, before_types: List[type], after_types: List[type]):
    module = astroid.parse(program)
    before, after = split_n_lines(module.body, init_lines)

    def compare(nodes: List[astns.NodeNG], node_types: List[type]):
        assert len(nodes) == len(node_types)

        for node, node_type in zip(nodes, node_types):
            assert isinstance(node, node_type)

    compare(before, before_types)
    compare(after, after_types)


@pytest.mark.parametrize("program,init_lines,modified", [
    (["x = 5"], 1, {"x": False}), # 0
    ([ # 1
        "x = 5",
        "x += 1"
    ], 1, {"x": True}),
    ([ # 2
        "x = 5",
        "x = 1"
    ], 1, {"x": True}),
    ([ # 3
        "x = 5",
        "y = 1",
        "x += 1"
    ], 2, {"x": True, "y": False}),
    ([ # 4
        "for x in range(10):",
        "    for x in range(11):",
        "        pass"
    ], 1, {"x": True}),
    ([ # 5
        "for x in range(10):",
        "    for y in range(11):",
        "        pass"
    ], 1, {"x": False}),
    ([ # 6
        "x = 0",
        "def foo():",
        "    global x",
        "    x = 1"
    ], 1, {"x": True}),
    ([ # 7
        "x = 0",
        "def foo():",
        "    global x",
        "    x += 1"
    ], 1, {"x": True}),
    ([ # 8
        "x = 0",
        "def foo():",
        "    nonlocal x",
        "    x = 1"
    ], 1, {"x": True}),
    ([ # 9
        "x = 0",
        "def foo():",
        "    x = 1",
        "    def bar():",
        "        nonlocal x",
        "        x = 2"
    ], 1, {"x": False}),
    ([ # 10
        "x = []",
        "def foo():",
        "    x.append(1)"
    ], 1, {"x": True}),
    ([ # 11
        "x = []",
        "def foo():",
        "    x[0] = 1"
    ], 1, {"x": True}),
    ([ # 12
        "x = None",
        "def foo():",
        "    x.y = 1"
    ], 1, {"x": True}),
    ([ # 13
        "x = None",
        "def foo():",
        "    x.y.z = 1"
    ], 1, {"x": True}),
    ([ # 14
        "x = None",
        "def foo():",
        "    x[0].y[1].z = 1"
    ], 1, {"x": True}),
    ([ # 15
        "x = None",
        "def foo():",
        "    x = []",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([ # 16
        "x = None",
        "def foo():",
        "    x, y = [], 0",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([ # 17
        "x = None",
        "def foo():",
        "    (x, y) = [], 0",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([ # 18
        "x = None",
        "def foo(x):",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([ # 19
        "x = None",
        "def foo(x: List[int] = []):",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([ # 20
        "x = 0",
        "if False:",
        "    x = 1",
    ], 1, {"x": True}),
    ([  # 21
        "x = 0",
        "while x < 10:",
        "    x += 1",
    ], 1, {"x": True}),
    ([  # 22
        "x = 0",
        "try:",
        "    y = 0",
        "except:",
        "    x = 1",
    ], 1, {"x": True}),
    ([  # 23
        "x = 0",
        "try:",
        "    y = 0",
        "except Exception:",
        "    x = 1",
    ], 1, {"x": True}),
    ([  # 24
        "x = 0",
        "try:",
        "    y = 0",
        "except Exception as e:",
        "    x = 1",
    ], 1, {"x": True}),
    ([  # 25
        "x = 0",
        "try:",
        "    y = 0",
        "except Exception as x:",
        "    y = 1",
    ], 1, {"x": True, "y": True}),
    ([  # 26
        "x = 0",
        "with open('f.txt') as x:",
        "    pass"
    ], 1, {"x": True}),
])
def test_var_event_listener(program: List[str], init_lines: int, modified: Dict[str, bool]):
    module = astroid.parse("\n".join(program))
    module.accept(CFGVisitor())

    def was_modified(in_varname: str) -> bool:
        in_scope = None
        for stmt in successors_from_loc(module.body[0].cfg_loc, explore_functions=True, include_start=True):
            for var, events in stmt.var_events.items():
                if in_varname != var.name:
                    continue
                for event in events:
                    if in_scope is None and event.type == VarEventType.ASSIGN:
                        in_scope = var.scope
                    if in_scope == var.scope and event.type in (VarEventType.REASSIGN, VarEventType.MODIFY):
                        return True
        return False

    VarModificationAnalysis.collect(module)
    for varname, mod in modified.items():
        assert was_modified(varname) == mod

    init, body = split_n_lines(module.body, 1)
    x_var = name_to_var("x", init[0])
    assert x_var is not None
    assert modified_in([x_var], body) == modified["x"]


def test_node_to_name_for_global_defined_after():
    program = [
        "def fun():",
        "    x = OUT",
        "OUT = 1"
    ]

    module = astroid.parse("\n".join(program))
    module.accept(CFGVisitor())
    VarModificationAnalysis.collect(module)

    assert name_to_var("OUT", module.body[0].body[0]) is not None
    assert node_to_var(module.body[0].body[0].value) is not None



from typing import Optional, Tuple
import os
import json
from dataclasses import dataclass
import time
from pathlib import Path

from platformdirs import PlatformDirs
from dataclasses_json import dataclass_json

from edulint.versions import pypi_helper

def current_timestamp() -> int:
    return int(time.time())


@dataclass_json
@dataclass
class PackageInfo:
    version: Optional[str] = None
    last_update_started: int = 0


class PackageInfoManager:
    @staticmethod
    def get_local_module_version(package_name: str) -> Optional[str]:
        try:
            from importlib.metadata import version  # This is only available in Python >= 3.8
            return version(package_name)
        except:
            pass

        try:
            import pkg_resources  # Part of setuptools, which might not be installed. But I don't want to include them as dependency.
            return pkg_resources.get_distribution(package_name).version
        except:
            pass

        return None

    @staticmethod
    def _create_json_file_if_doesnt_exist(filepath: str):
        if os.path.exists(filepath):
            return
        Path(filepath).parent.mkdir(parents=True, exist_ok=True)
        with open(filepath, "w", encoding="utf8") as f:
            f.write("{}")

    @classmethod
    def _package_info_path(cls):
        filepath = os.path.join(PlatformDirs(appname="edulint").user_data_dir, "versions.json")
        cls._create_json_file_if_doesnt_exist(filepath)
        return filepath 

    @classmethod
    def _save_package_info_locally(cls, package_name: str, version = None, last_update_started: Optional[int] = None) -> PackageInfo:
        with open(cls._package_info_path(), "r", encoding="utf8") as f:
            data = json.load(f)

        package_info = PackageInfo.from_dict(data.get(package_name, {}))

        if version:
            package_info.version = version
        if last_update_started:
            package_info.last_update_started = last_update_started
            
        data[package_name] = package_info.to_dict()
        with open(cls._package_info_path(), "w", encoding="utf8") as f:   # I'm aware there is possible race condition between read write. Risk accepted.
            json.dump(data, f, indent=4)
        
        return package_info

    @classmethod
    def _get_package_info_locally(cls, package_name: str) -> PackageInfo:
        with open(cls._package_info_path(), "r", encoding="utf8") as f:
            data = json.load(f)
            package_info = PackageInfo.from_dict(data.get(package_name, {}))
            return package_info

    @classmethod
    def get_latest_version(cls, package_name: str, ttl = 600) -> Optional[str]:
        package_info = cls._get_package_info_locally(package_name)
        if current_timestamp() < package_info.last_update_started + ttl:
            return package_info.version  # The might be None if the previous request didn't finish yet or if it failed.

        try:
            cls._save_package_info_locally(package_name, last_update_started=current_timestamp())
            versions = pypi_helper.get_versions(package_name)
            sorted_versions = list(sorted(versions, key = lambda x: (x.major, x.minor, x.micro)))
            latest_version_str = str(sorted_versions[-1])
            cls._save_package_info_locally(package_name, version=latest_version_str)
            return latest_version_str
        except Exception as _:
            return None

    @classmethod
    def is_update_waiting(cls, package_name: str, ttl = 600) -> bool:
        local_package_version = cls.get_local_module_version(package_name)
        if local_package_version is None: # unable to determine local package version
            return False

        latest_version = cls.get_latest_version(package_name, ttl)
        
        return local_package_version != latest_version  # This presumes local version is never higher than pypi version


if __name__ == "__main__":
    print("START")
    print(PackageInfoManager.get_latest_version("edulint"))
    print("END")



# Copied from https://github.com/GiraffeReversed/edulint-web/blob/main/setup.py

import requests
from typing import Dict, Any, List, Optional
from collections import defaultdict
import argparse

from edulint.versions.utils import Version

def _fully_released_versions(data: Dict[str, Any]) -> List[Version]:
    releases = data["releases"]

    version_ids = [v for v in releases.keys()]
    valid_versions: List[Version] = []

    for version_id in version_ids:
        has_some_builds: bool = bool(len(releases[version_id]))
        is_yanked: bool = any([x.get('yanked') for x in releases[version_id]])
        version_parsed: Optional[Version] = Version.parse(version_id)

        if has_some_builds and not is_yanked and version_parsed:
            valid_versions.append(version_parsed)
    
    return valid_versions


def _only_last_patch_of_each_minor(versions: List[Version]) -> List[Version]:
    major_minor: Dict[str, Version] = defaultdict(list)
    for version in versions:
        major_minor[f"{version.major}.{version.minor}"].append(version)
    for key in major_minor:
        major_minor[key].sort(reverse=True)
    patches_only = [major_minor[key][0] for key in major_minor]
    sorted_patches = list(sorted(patches_only, key = lambda x: (x.major, x.minor)))
    return sorted_patches


def get_versions(package_name: str = 'edulint') -> List[Version]:
    edulint_info = requests.get(f"https://pypi.org/pypi/{package_name}/json", timeout=3).json()
    version_ids: List[Version] = _fully_released_versions(edulint_info)
    if package_name == "pylint":
        version_ids = [version for version in version_ids if version.major >= 3]
    return version_ids



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('package_name')
    parser.add_argument('-g', '--github', default=False, action='store_true')
    parser.add_argument('-n', '--n-versions', type=int, default=5)
    args = parser.parse_args()
   

    package_name: str = args.package_name
    github_output: bool = args.github
    number_of_versions: int = args.n_versions

    versions = get_versions(package_name)
    versions = _only_last_patch_of_each_minor(versions)
    versions_str = [str(x) for x in versions]
    choosen_versions = versions_str[-number_of_versions:]
    
    if github_output:
        versions_as_str = str(choosen_versions).replace("'", "\"")
        # answer = f'{package_name}=\'{versions_as_str}\''
        
        # https://github.com/actions/runner/issues/1660#issuecomment-1359707506
        answer = f"""{package_name}<<EOF
{versions_as_str}
EOF"""
    else:
        answer = str(choosen_versions)
    
    print(answer)



FLAKE8 = {
    # pyflakes
    "F401": "%r imported but unused",
    "F402": "import %r from line %r shadowed by loop variable",
    "F403": "'from %s import *' used; unable to detect undefined names",
    "F404": "future import(s) name after other statements",
    "F405": "%r may be undefined, or defined from star imports: %s",
    "F406": "'from %s import *' only allowed at module level",
    "F407": "an undefined __future__ feature name was imported",
    "F501": "'...' %% ... has invalid format string: %s",
    "F502": "'...' %% ... expected mapping but got sequence",
    "F503": "'...' %% ... expected sequence but got mapping",
    "F504": "'...' %% ... has unused named argument(s): %s",
    "F505": "'...' %% ... is missing argument(s) for placeholder(s): %s",
    "F506": "'...' %% ... has mixed positional and named placeholders",
    "F507": "'...' %% ... has %d placeholder(s) but %d substitution(s)",
    "F508": "'...' %% ... `*` specifier requires sequence",
    "F509": "'...' %% ... has unsupported format character %r",
    "F521": "'...'.format(...) has invalid format string: %s",
    "F522": "'...'.format(...) has unused named argument(s): %s",
    "F523": "'...'.format(...) has unused arguments at position(s): %s",
    "F524": "'...'.format(...) is missing argument(s) for placeholder(s): %s",
    "F525": "'...'.format(...) mixes automatic and manual numbering",
    "F541": "f-string is missing placeholders",
    "F601": "dictionary key %r repeated with different values",
    "F602": "dictionary key variable %s repeated with different values",
    "F621": "too many expressions in an assignment with star-unpacking",
    "F622": "two or more starred expressions in an assignment (a, *b, *c = d)",
    "F631": "assertion test is a tuple, which is always True",
    "F632": "use ==/!= to compare constant literals (str, bytes, int, float, tuple)",
    "F633": "use of >> is invalid with print function",
    "F634": "if test is a tuple, which is always True",
    "F701": "a break statement outside of a while or for loop",
    "F702": "a continue statement outside of a while or for loop",
    "F704": "a yield or yield from statement outside of a function",
    "F706": "a return statement outside of a function/method",
    "F707": "an except: block as not the last exception handler",
    "F721": "syntax error in doctest",
    "F722": "syntax error in forward annotation %r",
    "F723": "syntax error in type comment",
    "F811": "redefinition of unused %r from line %r",
    "F821": "undefined name %r",
    "F822": "undefined name %r in __all__",
    # default = 'defined in enclosing scope on line %r'
    # builtin = 'defined as a builtin'
    "F823": "local variable %r %s referenced before assignment",
    "F831": "duplicate argument %r in function definition",
    "F841": "local variable %r is assigned to but never used",
    "F901": "'raise NotImplemented' should be 'raise NotImplementedError'",
    # pycodestyle
    "E101": "indentation contains mixed spaces and tabs",
    "E111": "indentation is not a multiple of %d",
    "E112": "expected an indented block",
    "W191": "indentation contains tabs",
    "W291": "trailing whitespace",
    "W293": "blank line contains whitespace",
    "W391": "blank line at end of file",
    "W292": "no newline at end of file",
    "E501": "line too long (%d > %d characters)",
    "E304": "blank lines found after function decorator",
    "E303": "too many blank lines (%d)",
    "E306": "expected %s blank line before a nested definition, found 0",
    "E301": "expected %s blank line, found 0",
    "E302": "expected %s blank lines, found %d",
    "E305": "expected %s blank lines after class or function definition, found %d",
    "E201": "whitespace after '%s'",
    "E274": "tab before keyword",
    "E272": "multiple spaces before keyword",
    "E273": "tab after keyword",
    "E271": "multiple spaces after keyword",
    "E275": "missing whitespace after keyword",
    "E124": "closing bracket does not match visual indentation",
    "E133": "closing bracket is missing indentation",
    "E128": "continuation line under-indented for visual indent",
    "E123": "closing bracket does not match indentation of opening bracket's line",
    "E122": "missing indentation or outdented",
    "E127": "over-indented for visual indent",
    "E131": "unaligned for hanging indent",
    "E126": "over-indented for hanging indent",
    "E121": "under-indented for hanging indent",
    "E129": "visually indented line",
    "E125": "continuation line",
    "E211": "whitespace before '%s'",
    "E223": "tab before operator",
    "E221": "multiple spaces before operator",
    "E224": "tab after operator",
    "E222": "multiple spaces after operator",
    "E225": "missing whitespace around operator",
    "E242": "tab after '%s'",
    "E241": "multiple spaces after '%s'",
    "E251": "unexpected spaces around keyword / parameter equals",
    "E252": "missing whitespace around parameter equals",
    "E261": "at least two spaces before inline comment",
    "E262": "inline comment should start with '# '",
    "E265": "block comment should start with '# '",
    "E266": "too many leading '#' for block comment",
    "E401": "multiple imports on one line",
    "E402": "module level import not at top of file",
    "E731": "do not assign a lambda expression, use a def",
    "E704": "multiple statements on one line (def)",
    "E701": "multiple statements on one line (colon)",
    "E702": "multiple statements on one line (semicolon)",
    "E703": "statement ends with a semicolon",
    "E502": "the backslash is redundant between brackets",
    "W503": "line break before binary operator",
    "W504": "line break after binary operator",
    "E712": "comparison to %s should be 'if%s cond:'",
    "E713": "test for membership should be 'not in'",
    "E714": "test for object identity should be 'is not'",
    "E721": "do not compare types, for exact checks use `is` / `is not`, for instance checks use `isinstance()`",
    "E722": "do not use bare 'except'",
    "E742": "ambiguous class definition '%s'",
    "E743": "ambiguous function definition '%s'",
    "E741": "ambiguous variable name '%s'",
    "W605": "invalid escape sequence '\\%s'",
    "W505": "doc line too long (%d > %d characters)",
}



import pytest
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from test_utils import lazy_problem, apply_and_lint, create_apply_and_lint
from typing import List


def _test_simplify_if(lines: List[str], expected_output: List[Problem], code: str) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, f"--enable={code}")],
        [p.set_code(code) for p in expected_output]
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def yyy(x):",
        "    if x:",
        "        return True",
        "    else:",
        "        return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x'")
    ]),
    ([
        "def xxx(x):",
        "    if x:",
        "        return False",
        "    else:",
        "        return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x>'"),
    ]),
    ([
        "def xxx(x):",
        "    if x:",
        "        return True",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x'"),
    ]),
    ([
        "def xxx(x):",
        "    if x:",
        "        return False",
        "    return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x>'"),
    ]),
])
def test_simplify_if_statement_single_var_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6201")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return True",
        "    return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return False",
        "    return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return y",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return y",
        "    return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return True",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return False",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return y",
        "    else:",
        "        return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return y",
        "    else:",
        "        return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> or y'")
    ]),
])
def test_simplify_if_statement_two_vars_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6202")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return y",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and z and y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x or z:",
        "        return y",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return (x or z) and y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        return y and z",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and y and z'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        return y or z",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x and (y or z)'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return True",
        "    return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return (x and z) or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return False",
        "    return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated (x and z)> and y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        return False",
        "    return y or z"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated x> and (y or z)'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x or z:",
        "        return y",
        "    return False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return (x or z) and y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return y",
        "    return True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return <negated (x and z)> or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and z:",
        "        return True",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return (x and z) or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x or z:",
        "        return True",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x or z or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        return True",
        "    else:",
        "        return y and z"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be replaced with 'return x or (y and z)'")
    ]),
])
def test_simplify_if_statement_three_vars_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6202")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        if y:",
        "            return 0",
        "    return 1"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be merged with the nested one to 'if x and y:'")
    ]),
])
def test_simplify_if_statement_nested_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6207")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return 0",
        "    if y:",
        "        return 0",
        "    return 1"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if statement can be merged with the following one to 'if x or y:'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        return True",
        "    if y:",
        "        return False",
        "    return True"
    ], [
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x and y:",
        "        return True",
        "    if z:",
        "        return True",
        "    return True"
    ], [
    ]),
])
def test_simplify_if_statement_multiconds_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6208")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def is_right(a, b, c):",
        "    if c ** 2 == a ** 2 + b ** 2 or a ** 2 == c ** 2 + b ** 2 or \\",
        "       b ** 2 == a ** 2 + c ** 2:",
        "        triangle_is_righ = True",
        "    else:",
        "        triangle_is_righ = False",
        "    return triangle_is_righ"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replaced with 'triangle_is_righ = c**2 == a**2 "
                  "+ b**2 or a**2 == c**2 + b**2 or b**2 == a**2 + c**2'")
    ]),
])
def test_simplify_if_assigning_statement_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6203")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        a = True",
        "    else:",
        "        a = y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replaced with 'a = x or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        a = False",
        "    else:",
        "        a = y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replaced with 'a = <negated x> and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        a = y",
        "    else:",
        "        a = False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replaced with 'a = x and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    if x:",
        "        a = y",
        "    else:",
        "        a = True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replaced with 'a = <negated x> or y'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        a = y and z",
        "    else:",
        "        a = True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replaced with 'a = <negated x> or (y and z)'")
    ]),
    ([
        "def xxx(x, y, z):",
        "    if x:",
        "        a = y or z",
        "    else:",
        "        a = True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The conditional assignment can be replaced with 'a = <negated x> or y or z'")
    ]),
])
def test_simplify_if_assigning_statement_conj_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6210")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "report = []",
        "which = 0",
        "report[which] = True if report[which] > \\",
        "    report[which] else False"
    ], [
        lazy_problem().set_line(3)
        .set_text("The if expression can be replaced with 'report[which] > report[which]'")
    ]),
    ([
        "report = []",
        "which = 0",
        "report[which], x = True if report[which] > report[which] else False, 0"
    ], [
        lazy_problem().set_line(3)
        .set_text("The if expression can be replaced with 'report[which] > report[which]'")
    ]),
    ([
        "report = []",
        "which = 0",
        "report[which], x = True if report[which] > report[which] else False, \\",
        "    False if report[which] <= report[which] else True"
    ], [
        lazy_problem().set_line(3)
        .set_text("The if expression can be replaced with 'report[which] > report[which]'"),
        lazy_problem().set_line(4)
        .set_text("The if expression can be replaced with '<negated report[which] <= report[which]>'")
    ]),
])
def test_simplify_if_expression_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6204")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    r = True if x else y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if expression can be replaced with 'x or y'")
    ]),
    ([
        "def xxx(x, y):",
        "    r = False if x else y"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if expression can be replaced with '<negated x> and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    r = y if x else False"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if expression can be replaced with 'x and y'")
    ]),
    ([
        "def xxx(x, y):",
        "    r = y if x else True"
    ], [
        lazy_problem().set_line(2)
        .set_text("The if expression can be replaced with '<negated x> or y'")
    ]),
])
def test_simplify_if_expression_conj_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6209")


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x, y):",
        "    if x:",
        "        pass",
        "    else:",
        "        return y"
    ], [
        lazy_problem().set_line(2)
        .set_text("Use 'if <negated x>: <else body>' instead of 'pass'")
    ]),
])
def test_simplify_if_pass_custom(lines: List[str], expected_output: List[Problem]) -> None:
    _test_simplify_if(lines, expected_output, "R6205")


@pytest.mark.parametrize("filename,expected_output", [
    ("015080-p4_geometry.py", [
        lazy_problem().set_code("R6207").set_line(14)
        .set_text("The if statement can be merged with the nested one to 'if len(sides) != len(set(sides))"
                  " and len(set(sides)) <= 3 and sides[1] == sides[2]:'"),
        lazy_problem().set_code("R6201").set_line(21)
        .set_text("The if statement can be replaced with 'return side_c == sides[2]'"),
        lazy_problem().set_code("R6201").set_line(32)
        .set_text("The if statement can be replaced with 'return a == b & a == c'"),
        lazy_problem().set_code("R6201").set_line(47)
        .set_text("The if statement can be replaced with 'return <negated duplicate(sides) < max(sides)>'"),
    ]),
    ("z202817-zkouska.py", [
        lazy_problem().set_code("R6208").set_line(76)
        .set_text("The if statement can be merged with the following one to "
                  "'if word == '' or word[len(word) - 1] == '.':'"),
        lazy_problem().set_code("R6205").set_line(82),
    ]),
    ("014823-p4_geometry.py", [
            lazy_problem().set_code("R6206").set_line(17)
            .set_text("Both branches should return a value explicitly (one returns implicit None)"),
            lazy_problem().set_code("R6206").set_line(25),
            lazy_problem().set_code("R6206").set_line(33),
            lazy_problem().set_code("R6201").set_line(43),
            lazy_problem().set_code("R6201").set_line(55),
    ]),
    ("024423-p5_credit.py", []),
    ("044669-p3_person_id.py", [lazy_problem().set_code("R6206").set_line(30)])
])
def test_simplify_if_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=simplifiable-if")],
        expected_output
    )

@pytest.mark.parametrize("lines,expected_output", [
    ([
        "x = 7.8",
        "y = [-5.25, 2.3, 3, 4]",
        "z = -20",
        "a = abs(min(y)) > -4.47 and abs(min(y)) < 4.47",
        "c = int(z + abs(sum(reversed(sorted(y))))) < 8.56 and -8.56 < int(z + abs(sum(reversed(sorted(y)))))",
        "d = x + len(y) < 5 and x < 5 and x + len(y) > -10 and x > -5 and x + len(y) > -5",
    ], [
        lazy_problem().set_line(4)
        .set_text("'abs(min(y)) > -4.47 and abs(min(y)) < 4.47' can be replaced with 'abs(min(y)) < 4.47'"),
        lazy_problem().set_line(5)
        .set_text("'int(z + abs(sum(reversed(sorted(y))))) < 8.56 and int(z + abs(sum(reversed(sorted(y))))) > -8.56' can be replaced with 'abs(int(z + abs(sum(reversed(sorted(y)))))) < 8.56'"),
        lazy_problem().set_line(6)
        .set_text("'x + len(y) < 5 and x + len(y) > -10 and x + len(y) > -5' can be replaced with 'abs(x + len(y)) < 5'"),
        lazy_problem().set_line(6)
        .set_text("'x < 5 and x > -5' can be replaced with 'abs(x) < 5'"),
    ]),
    ([
        "x = 7.8",
        "y = [-5.25, 2.3, 3, 4]",
        "a = x >= 2 or x <= -2",
        "b = x + len(y) > 4 or x + len(y) < -4",
        "c = x + 1 < -9 or 9 < x + 1",
        "d = x > 5 or x > 4 or x <= -7 or x >= 3 or x < -5 or x <= -3",
        "e = -0.5 <= len(y) and len(y) < 5 and -7 <= x or x > 7 or x < -7",
        "f = y[0] >= 5 or y[0] <= -5",
    ], [
        lazy_problem().set_line(3)
        .set_text("'x >= 2 or x <= -2' can be replaced with 'abs(x) >= 2'"),
        lazy_problem().set_line(4)
        .set_text("'x + len(y) > 4 or x + len(y) < -4' can be replaced with 'abs(x + len(y)) > 4'"),
        lazy_problem().set_line(5)
        .set_text("'x + 1 < -9 or x + 1 > 9' can be replaced with 'abs(x + 1) > 9'"),
        lazy_problem().set_line(6)
        .set_text("'x > 5 or x > 4 or x <= -7 or x >= 3 or x < -5 or x <= -3' can be replaced with 'abs(x) >= 3'"),
        lazy_problem().set_line(7)
        .set_text("'x > 7 or x < -7' can be replaced with 'abs(x) > 7'"),
        lazy_problem().set_line(8)
        .set_text("'y[0] >= 5 or y[0] <= -5' can be replaced with 'abs(y[0]) >= 5'"),
    ]),
])
def test_simplifiable_with_abs_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=simplifiable-with-abs")],
        expected_output,
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("ffed796f7c-gps.py", [
        lazy_problem().set_line(23)
        .set_text("'float(y[1]) > 90 or float(y[1]) < -90 or float(y[1]) > 180 or float(y[1]) < -180' can be replaced with 'abs(float(y[1])) > 90'")
    ]),
    ("ffed796f7c-gps_tweaked.py", [
        lazy_problem().set_line(23)
        .set_text("'float(y[1]) > 90 or float(y[1]) < -90' can be replaced with 'abs(float(y[1])) > 90'"),
        lazy_problem().set_line(23)
        .set_text("'float(y[2]) > 180 or float(y[2]) < -180' can be replaced with 'abs(float(y[2])) > 180'")
    ])
])
def test_simplifiable_with_abs(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=simplifiable-with-abs")],
        expected_output,
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "x = 7.8",
        "y = [-5.25, 2.3, 3, 4]",
        "a = x > -2 or x <= 6",
        "c = x >= 1 or x > 1",
        "d = x < 2 and x <= -5",
        "e = -(x - 1) > 5 or - (x- 1) < -5 or -(x-1) >= 4",
    ], [
        lazy_problem().set_line(3)
        .set_text("'x > -2 or x <= 6' can be simplified to 'True'. Simplify the condition if it is on purpose, or change it if it was not."),
        lazy_problem().set_line(4)
        .set_text("'x >= 1 or x > 1' can be simplified to 'x >= 1'. Simplify the condition if it is on purpose, or change it if it was not."),
        lazy_problem().set_line(5)
        .set_text("'x < 2 and x <= -5' can be simplified to 'x <= -5'. Simplify the condition if it is on purpose, or change it if it was not."),
        lazy_problem().set_line(6)
        .set_text("'-(x - 1) > 5 or -(x - 1) < -5 or -(x - 1) >= 4' can be simplified to '-(x - 1) < -5 or -(x - 1) >= 4'. Simplify the condition if it is on purpose, or change it if it was not."),
    ]),
    ([
        "class Person:",
        "    def __init__(self, name, age):",
        "        self.name = name",
        "        self.age = age",
        "    def this_not_pure(self, something):",
        "        something.append(5)",
        "        return something[0]",
        "def foo():",
        "    person = Person('Albert', 56)",
        "    a = person.age > 5 and person.age > 9",
        "    b = person.this_not_pure(y) > 6 or person.this_not_pure(y) < -6",
    ], [
        lazy_problem().set_line(10)
        .set_text("'person.age > 5 and person.age > 9' can be simplified to 'person.age > 9'. Simplify the condition if it is on purpose, or change it if it was not."),
    ]),
])
def test_redundant_compare(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=redundant-compare-in-condition")],
        expected_output,
    )

@pytest.mark.parametrize("lines,expected_output", [
    ([
        "x = 7.8",
        "b = x > 4 or x > 0 or x < -1 or x < 0",
        "c = x >= 9 and x <= 9"
    ], [
        lazy_problem().set_line(2)
        .set_text("'x > 4 or x > 0 or x < -1 or x < 0' can be replaced with 'x != 0'"),
        lazy_problem().set_line(3)
        .set_text("'x >= 9 and x <= 9' can be replaced with 'x == 9'"),
    ]),
])
def test_using_compare_instead_of_equal(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=using-compare-instead-of-equal")],
        expected_output,
    )

@pytest.mark.parametrize("lines,expected_output", [
    ([
        "a < b and a+50 < b",
        "j >= n//2 and i>=n//2",
        "i < len(t1) or len(t2) > i or i < 5 or i < max( x,y , z )",
        "- 1 + len(lst) > min(x, 3, y) or max(y, z) < -1 + len(lst) or (-1 + len(lst)) > min(x, z)"
    ], [
        lazy_problem().set_line(1)
        .set_text("'a < b and a + 50 < b' can be replaced with 'max(a, a + 50) < b'"),
        lazy_problem().set_line(2)
        .set_text("'j >= n // 2 and i >= n // 2' can be replaced with 'min(j, i) >= n // 2'"),
        lazy_problem().set_line(3)
        .set_text("'i < len(t1) or len(t2) > i or i < max(x, y, z) or i < 5' can be replaced with 'i < max(len(t1), len(t2), x, y, z, 5)'"),
        lazy_problem().set_line(4)
        .set_text("'-1 + len(lst) > min(x, 3, y) or max(y, z) < -1 + len(lst) or -1 + len(lst) > min(x, z)' can be replaced with 'min(x, 3, y, max(y, z), x, z) < -1 + len(lst)'"),
    ]),
])
def test_redundant_compare_with_variables(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=redundant-compare-avoidable-with-max-min")],
        expected_output,
    )

@pytest.mark.parametrize("lines,expected_output", [
    ([
        "x = 1",
        "y = 7",
        "lst = [2, 3, 0, 8]",
        "x % 2 != 1 and y % 2 != 1 or x % 2 == 1 and y % 2 == 1",
        "x % 4 == 0 and y % 4 == 0 or x % 4 == 1 and y % 4 == 1 or x % 4 != 0 and y % 4 != 0 or x % 4 == 2 and y % 4 == 2 or x % 4 == 3 and y % 4 == 3",
        "len(lst) % 2 == 0 and (y - 2) % 2 == 0 or 1 == (y - 2) % 2 and len(lst) % 2 == 1 or (y - 2) % 2 == 0 and 0 == len(lst) % 2",
        "(x % 3 == 0 and 0 != y % 3) or (y % 3 == 0 and x % 3 != 0) or (x % 3 != 1 and y % 3 == 1) or (x % 3 == 1 and y % 3 == 2)",
        "x % 5 != 1 and y % 5 != 3 or x % 5 != 1 and y % 5 == 3 or x % 5 != 2 and y % 5 != 4",
        "x % 5 != 1 and y % 5 != 3 or x % 5 != 1 and y % 5 == 3 or x % 5 == 1 and y % 5 != 3 or x % 5 != 2 and y % 5 != 4",
        "x % 2 != 0 and y % 2 == 0 or 0 == x % 3 and y % 3 == 0 or y % 3 == 1 and x % 3 == 1 or x % 2 == 0 and y % 2 == 1 or x % 2 != 1 and y % 2 == 1 or y % 3 == 2 and x % 3 == 2"
    ], [
        lazy_problem().set_line(4)
        .set_text("'x % 2 != 1 and y % 2 != 1 or x % 2 == 1 and y % 2 == 1' can be replaced with 'x % 2 == y % 2'"),
        lazy_problem().set_line(5)
        .set_text("'x % 4 == 0 and y % 4 == 0 or x % 4 == 1 and y % 4 == 1 or x % 4 == 2 and y % 4 == 2 or x % 4 == 3 and y % 4 == 3' can be replaced with 'x % 4 == y % 4'"),
        lazy_problem().set_line(6)
        .set_text("'len(lst) % 2 == 0 and (y - 2) % 2 == 0 or 1 == (y - 2) % 2 and len(lst) % 2 == 1 or (y - 2) % 2 == 0 and 0 == len(lst) % 2' can be replaced with 'len(lst) % 2 == (y - 2) % 2'"),
        lazy_problem().set_line(7)
        .set_text("'x % 3 == 0 and 0 != y % 3 or y % 3 == 0 and x % 3 != 0 or x % 3 != 1 and y % 3 == 1 or x % 3 == 1 and y % 3 == 2' can be replaced with 'x % 3 != y % 3'"),
        lazy_problem().set_line(9)
        .set_text("'x % 5 != 1 and y % 5 != 3 or x % 5 != 1 and y % 5 == 3 or x % 5 == 1 and y % 5 != 3 or x % 5 != 2 and y % 5 != 4' can be replaced with 'True'"),
        lazy_problem().set_line(10)
        .set_text("'x % 2 != 0 and y % 2 == 0 or x % 2 == 0 and y % 2 == 1 or x % 2 != 1 and y % 2 == 1' can be replaced with 'x % 2 != y % 2'"),
        lazy_problem().set_line(10)
        .set_text("'0 == x % 3 and y % 3 == 0 or y % 3 == 1 and x % 3 == 1 or y % 3 == 2 and x % 3 == 2' can be replaced with 'x % 3 == y % 3'"),
    ]),
])
def test_simplifiable_test_by_equals(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=simplifiable-test-by-equals")],
        expected_output,
    )



from collections import namedtuple
from typing import Tuple, List, Generator

from astroid import nodes

from edulint.linting.analyses.antiunify import AunifyVar, new_node, cprint  # noqa: F401


Fixed = namedtuple("Fixed", ["symbol", "tokens", "statements", "message_args"])


### antiunification conditions


def length_mismatch(avars) -> bool:
    for avar in avars:
        some = avar.subs[0]
        if not isinstance(some, list) and any(isinstance(sub, list) for sub in avar.subs):
            return True
        if isinstance(some, list) and any(len(sub) != len(some) for sub in avar.subs):
            return True
    return False


def type_mismatch(avars, allowed_mismatches=None) -> bool:
    allowed_mismatches = allowed_mismatches if allowed_mismatches is not None else []
    for avar in avars:
        sub_types = {type(sub) for sub in avar.subs}
        if any(sub_types.issubset(am) for am in allowed_mismatches):
            continue
        if len(sub_types) > 1:
            return True
    return False


def called_aunify_var(avars) -> bool:
    for avar in avars:
        node = avar.parent
        if (
            (isinstance(node, nodes.Compare) and avar in [o for o, n in node.ops])
            or (isinstance(node, nodes.BinOp) and avar == node.op)
            or (isinstance(node, nodes.AugAssign) and avar == node.op)
            or (isinstance(node, nodes.Attribute) and avar == node.attrname)
        ):
            return True

        while node is not None:
            if isinstance(node.parent, nodes.Call) and node == node.parent.func:
                return True
            node = node.parent
    return False


def assignment_to_aunify_var(avars) -> bool:
    return any(isinstance(avar.parent, (nodes.AssignName, nodes.AssignAttr)) for avar in avars)


### general


def is_duplication_candidate(stmtss) -> bool:
    for ns in zip(*stmtss):
        if not all(isinstance(n, type(ns[0])) for n in ns):
            return False
    return True


def saves_enough_tokens(tokens_before: int, stmts_before: int, fixed: Fixed):
    if fixed.symbol in ("similar-if-to-untwisted", "similar-if-into-block", "similar-if-to-use"):
        return fixed.statements <= stmts_before and fixed.tokens < tokens_before
    return fixed.statements <= stmts_before + 1 and fixed.tokens < 0.8 * tokens_before


def get_loop_repetitions(
    block: List[nodes.NodeNG],
) -> Generator[Tuple[int, List[List[nodes.NodeNG]]], None, None]:
    for end in range(len(block), 0, -1):
        for subblock_len in range(1, end // 2 + 1):
            if end % subblock_len != 0:
                continue
            subblocks = [block[i : i + subblock_len] for i in range(0, end, subblock_len)]
            yield ((end // subblock_len) * subblock_len, subblocks)


def to_node(val, avar=None) -> nodes.NodeNG:
    assert not isinstance(val, list)
    if isinstance(val, nodes.NodeNG):
        return val
    if avar is not None and isinstance(avar.parent, nodes.Name):
        return new_node(nodes.Name, name=val)
    return new_node(nodes.Const, value=val)


def to_parent(val: AunifyVar) -> nodes.NodeNG:
    parent = val.parent
    if isinstance(parent, (nodes.Const, nodes.Name)):
        parent = parent.parent
    assert parent is not None
    return parent


def get_common_parent(ns: List[nodes.NodeNG]) -> bool:
    if len(ns) == 0:
        return None

    if len(ns) == 1:
        return to_parent(ns[0])

    fst_parents = [ns[0]] + list(ns[0].node_ancestors())
    other_parents = set.intersection(
        *[{ns[i]} | set(ns[i].node_ancestors()) for i in range(1, len(ns))]
    )

    for parent in fst_parents:
        if parent in other_parents:
            return parent
    return None


def to_start_lines(to_aunify: List[List[nodes.NodeNG]]) -> str:
    return ", ".join(str(n[0].fromlineno) for n in to_aunify)



from edulint.linters import Linter
from edulint.options import Option
from edulint.config.arg import ImmutableArg
from edulint.linting.problem import Problem
from dataclasses import dataclass
from typing import Callable, Optional, Pattern, Match, AnyStr, Dict, Tuple, Set, List
import re


@dataclass
class Tweaker:
    used_options: Set[Option]

    pattern: Pattern[AnyStr]  # type: ignore
    keep: Callable[["Tweaker", Problem, List[ImmutableArg]], bool] = lambda _t, _p, _a: True
    reword: Optional[Callable[["Tweaker", Problem], str]] = None

    def match(self, problem: Problem) -> Match[str]:
        match = self.pattern.match(problem.text)
        assert match is not None
        return match

    def should_keep(self, problem: Problem, args: List[ImmutableArg]) -> bool:
        return self.keep(self, problem, args)

    def does_reword(self) -> bool:
        return self.reword is not None

    def get_reword(self, problem: Problem) -> str:
        return self.reword(self, problem) if self.reword else problem.text


def disallowed_name_keep(self: Tweaker, problem: Problem, args: List[ImmutableArg]) -> bool:
    if (
        len(args) == 1
        and args[0].option == Option.ALLOWED_ONECHAR_NAMES
        and args[0].val is not None
    ):
        assert isinstance(args[0].val, str)
        allowed_onechar_names = args[0].val
    else:
        allowed_onechar_names = ""

    match = self.match(problem)
    name = match.group(1)

    return len(name) != 1 or name not in allowed_onechar_names


def disallowed_name_reword(self: Tweaker, problem: Problem) -> str:
    name = self.match(problem).group(1)
    if len(name) == 1:
        return f'Disallowed single-character variable name "{name}", choose a more descriptive name'
    return problem.text


def consider_using_in_reword(self: Tweaker, problem: Problem) -> str:
    match = self.match(problem)

    groups = match.groups()
    start = groups[0]
    outer_quote = groups[2]
    vals = groups[5].split(", ")
    assert vals

    if all(val.count("(") == val.count(")") for val in vals) and all(
        v and v[0] == v[-1] and v[0] in "\"'" and len(v) == 3 for v in vals
    ):
        inner_quote = vals[0][0]
        return (
            start + inner_quote + "".join(v.strip("\"'") for v in vals) + inner_quote + outer_quote
        )

    return start + f"({', '.join(vals)})" + outer_quote


def unused_import_keep(self: Tweaker, problem: Problem, args: List[ImmutableArg]) -> bool:
    match = self.match(problem)
    return not match.group(1).startswith("ib111")


def no_repeated_op_keep(self: Tweaker, problem: Problem, args: List[ImmutableArg]) -> bool:
    assert len(args) == 1

    if "enhancement" in args[0].val:
        return True

    match = self.match(problem)
    expr = match.group(1)
    return bool(args[0].val) or expr.count("*") > 1


def superfluous_parens_keep(self: Tweaker, problem: Problem, args: List[ImmutableArg]) -> bool:
    match = self.match(problem)
    keyword = match.group(1).strip("'\"")

    return keyword.lower() != "not"


def redefined_builtin_keep(self: Tweaker, problem: Problem, args: List[ImmutableArg]) -> bool:
    match = self.match(problem)
    redefined_builtin = match.group(1).strip("'\"")

    assert len(args) == 1 and args[0].option == Option.DISALLOWED_BUILTIN_NAMES
    disallowed_builtin_names = args[0].val

    return not disallowed_builtin_names or redefined_builtin in disallowed_builtin_names


def singleton_bool_comparison_reword(self: Tweaker, problem: Problem) -> bool:
    match = self.match(problem)
    return "".join(match.groups())


Tweakers = Dict[Tuple[Linter, str], Tweaker]

TWEAKERS = {
    (Linter.PYLINT, "C0104"): Tweaker(  # disallowed-name
        set([Option.ALLOWED_ONECHAR_NAMES]),
        re.compile(r"Disallowed name \"(.*)\""),
        disallowed_name_keep,
        disallowed_name_reword,
    ),
    # (Linter.PYLINT, "R1714"): Tweaker(  # consider-using-in
    #     set(),
    #     re.compile(
    #         r"^(Consider merging these comparisons with ['\"]in['\"] (to|by using) "
    #         r"(\"|\')([^\s]*)( not)? in )\((.+)\)(\"|\')"
    #     ),
    #     reword=consider_using_in_reword,
    # ),
    (Linter.FLAKE8, "F401"): Tweaker(  # module imported but unused
        set(), re.compile("'(.*)' imported but unused"), unused_import_keep
    ),
    (Linter.PYLINT, "R6607"): Tweaker(  # no-repeated-op
        set([Option.SET_GROUPS]),
        re.compile(r"^Use [^ ]* instead of repeated [^ ]* in ([^\.]*)."),
        no_repeated_op_keep,
    ),
    (Linter.PYLINT, "C0325"): Tweaker(  # superfluous-parens
        set(),
        re.compile(r"Unnecessary parens after (.*) keyword"),
        superfluous_parens_keep,
    ),
    (Linter.PYLINT, "W0622"): Tweaker(  # redefined-builtin
        set([Option.DISALLOWED_BUILTIN_NAMES]),
        re.compile(r"^Redefining built-in (.*)"),
        redefined_builtin_keep,
    ),
    (Linter.FLAKE8, "E712"): Tweaker(  # singleton bool comparison
        set(),
        re.compile(r"(comparison to (?:True|False) should be )'if cond is .*' or (.*)"),
        reword=singleton_bool_comparison_reword,
    ),
}


def get_tweakers() -> Tweakers:
    return TWEAKERS



from typing import List

from astroid import nodes  # type: ignore

from edulint.linting.analyses.antiunify import antiunify, cprint  # noqa: F401
from edulint.linting.analyses.data_dependency import (
    get_vars_defined_before,
    get_vars_used_after,
    get_control_statements,
)
from edulint.linting.checkers.utils import (
    get_statements_count,
    get_token_count,
    EXPRESSION_TYPES,
    new_node,
)
from edulint.linting.checkers.duplication.utils import (
    Fixed,
    length_mismatch,
    type_mismatch,
    called_aunify_var,
    assignment_to_aunify_var,
    saves_enough_tokens,
    to_node,
    to_start_lines,
)

### similar to function


def get_fixed_by_function(to_aunify, core, avars):
    # compute necessary arguments from different values
    seen = {}
    for avar in avars:
        var_vals = tuple(avar.subs)
        old_avar = seen.get(var_vals, avar)

        if old_avar != avar:
            continue
        seen[var_vals] = avar

    # compute extras
    extra_args = get_vars_defined_before(core)
    return_vals_needed = len(get_vars_used_after(core))
    control_needed = len(get_control_statements(core))

    # generate calls in ifs
    calls = []
    for i in range(len(to_aunify)):
        params = [s[i] for s in seen]
        call = new_node(
            nodes.Call,
            func=new_node(nodes.Name, name="AUX"),
            args=[to_node(param) for param in params]
            + [new_node(nodes.Name, name=var.name) for var in extra_args.keys()],
        )
        if return_vals_needed + control_needed == 0:
            calls.append(call)
        else:
            assign = new_node(
                nodes.Assign,
                targets=[
                    new_node(nodes.AssignName, name=f"<r{i}>")
                    for i in range(control_needed + return_vals_needed)
                ],
                value=call,
            )
            calls.append(assign)

            # generate management for returned control flow
            for i in range(control_needed):
                test = new_node(
                    nodes.BinOp,
                    op="is",
                    left=new_node(nodes.Name, name=f"<r{i}>"),
                    right=new_node(nodes.Const, value=None),
                )
                if_ = new_node(
                    nodes.If, test=test, body=[new_node(nodes.Return)]
                )  # placeholder for a control
                calls.append(if_)

    # generate function
    fun_def = new_node(
        nodes.FunctionDef,
        name="AUX",
        args=new_node(
            nodes.Arguments,
            args=[new_node(nodes.AssignName, name=avar.name) for avar in seen.values()]
            + [new_node(nodes.AssignName, name=var.name) for var in extra_args.keys()],
        ),
        body=core if isinstance(core, list) else [core],
    )

    return Fixed(
        "similar-block-to-function",
        get_token_count(calls) + get_token_count(fun_def),
        get_statements_count(calls, include_defs=False, include_name_main=True)
        + get_statements_count(fun_def, include_defs=False, include_name_main=True),
        (
            len(to_aunify),
            get_statements_count(core, include_defs=False, include_name_main=True),
            to_start_lines(to_aunify),
        ),
    )


def similar_to_function(self, to_aunify: List[List[nodes.NodeNG]], core, avars) -> bool:
    if type_mismatch(avars):
        return False

    tokens_before = sum(get_token_count(node) for node in to_aunify)
    stmts_before = sum(
        get_statements_count(node, include_defs=False, include_name_main=True) for node in to_aunify
    )

    fixed = get_fixed_by_function(to_aunify, core, avars)
    if not saves_enough_tokens(tokens_before, stmts_before, fixed):
        return False

    message_id, _tokens, _statements, message_args = fixed

    first = to_aunify[0][0]
    last = to_aunify[0][-1]
    self.add_message(
        message_id,
        line=first.fromlineno,
        col_offset=first.col_offset,
        end_lineno=last.tolineno,
        end_col_offset=last.col_offset,
        args=message_args,
    )
    return True


### similar to call


def is_possible_callee(function: nodes.FunctionDef, sub_aunify):
    assert len(sub_aunify) <= len(function.body)
    if len(sub_aunify) == len(function.body):
        return True

    return sub_aunify == function.body[: len(sub_aunify)] and isinstance(
        function.body[len(sub_aunify)], nodes.Return
    )


def get_possible_callees(to_aunify):
    possible_callees = []

    for i, sub_aunify in enumerate(to_aunify):
        if not isinstance(sub_aunify[0].parent, nodes.FunctionDef):
            continue
        function = sub_aunify[0].parent
        if is_possible_callee(function, sub_aunify):
            possible_callees.append((i, function))

    return possible_callees


def returns_used_value(return_, returned_values, node):
    # TODO do properly, not by checking for substring
    if node == return_ or any(r in node.as_string() for r in returned_values):
        return True

    for parent in node.node_ancestors():
        if parent == return_:
            return True
        if any(r in parent.as_string() for r in returned_values):
            return True

    return False


def similar_to_call(self, to_aunify: List[List[nodes.NodeNG]], core, avars) -> bool:
    possible_callees = get_possible_callees(to_aunify)
    if len(possible_callees) != 1:
        return False

    i, function = possible_callees[0]
    args = function.args.arguments
    argnames = {arg.name for arg in args}

    for avar in avars:
        sub = avar.subs[i]
        if not isinstance(sub, nodes.Name) or sub.name not in argnames:
            return False

    vars_used_after = get_vars_used_after(core)
    if len(vars_used_after) != 0:
        if len(function.body) <= len(to_aunify[i]):
            return False
        last = function.body[len(to_aunify[i])]  # handle unreachable code
        if not isinstance(last, nodes.Return) or last.value is None:
            return False

        returned_values = (
            [last.value.as_string()]
            if not isinstance(last.value, nodes.Tuple)
            else [e.as_string() for e in last.value.elts]
        )
        for users in vars_used_after.values():
            for node in users:
                if not returns_used_value(last, returned_values, node):
                    return False

    other_body = to_aunify[0] if i != 0 else to_aunify[1]
    first = other_body[0]
    last = other_body[-1]
    self.add_message(
        "similar-block-to-call",
        line=first.fromlineno,
        col_offset=first.col_offset,
        end_lineno=last.tolineno,
        end_col_offset=last.col_offset,
        args=(function.name),
    )
    return True


### control function


def similar_to_block(checker, to_aunify: List[List[nodes.NodeNG]]) -> bool:
    result = antiunify(
        to_aunify,
        stop_on=lambda avars: length_mismatch(avars)
        or type_mismatch(avars, allowed_mismatches=[{nodes.Name, t} for t in EXPRESSION_TYPES])
        or called_aunify_var(avars),
        stop_on_after_renamed_identical=lambda avars: assignment_to_aunify_var(avars),
    )
    if result is None:
        return False
    core, avars = result

    if all(isinstance(vals[0], nodes.FunctionDef) for vals in to_aunify):
        return False  # TODO hint use common helper function

    if similar_to_call(checker, to_aunify, core, avars):
        return True
    return similar_to_function(checker, to_aunify, core, avars)
